import logging
import pathlib
from unittest import mock

import boost_histogram as bh
import numpy as np
import pytest

from cabinetry import template_builder


def test__check_for_override():
    # override exists for template
    assert (
        template_builder._check_for_override(
            {"Up": {"setting": "val"}}, "Up", "setting"
        )
        == "val"
    )

    # no override for template
    assert template_builder._check_for_override({}, "Up", "setting") is None

    # no option requested
    assert (
        template_builder._check_for_override({"Up": {"setting": "val"}}, "Up", "")
        is None
    )

    # no template requested
    assert (
        template_builder._check_for_override({"Up": {"setting": "val"}}, "", "setting")
        is None
    )

    # override is a list
    assert template_builder._check_for_override(
        {"Up": {"setting": ["val", "val2"]}}, "Up", "setting"
    ) == ["val", "val2"]


def test__get_ntuple_paths(caplog):
    # only general path, no override
    assert template_builder._get_ntuple_paths("path.root", {}, {}, {}, "") == [
        pathlib.Path("path.root")
    ]

    # general path with region and sample templates
    assert (
        template_builder._get_ntuple_paths(
            "{RegionPath}/{SamplePaths}",
            {"RegionPath": "region"},
            {"SamplePaths": "sample.root"},
            {},
            "",
        )
        == [pathlib.Path("region/sample.root")]
    )

    # two SamplePaths
    assert (
        template_builder._get_ntuple_paths(
            "{RegionPath}/{SamplePaths}",
            {"RegionPath": "region"},
            {"SamplePaths": ["sample.root", "new.root"]},
            {},
            "",
        )
        == [pathlib.Path("region/sample.root"), pathlib.Path("region/new.root")]
    )

    # systematic with override for RegionPath and SamplePaths
    assert (
        template_builder._get_ntuple_paths(
            "{RegionPath}/{SamplePaths}",
            {"RegionPath": "reg_1"},
            {"SamplePaths": "path.root"},
            {
                "Name": "variation",
                "Up": {
                    "SamplePaths": ["variation.root", "new.root"],
                    "RegionPath": "reg_2",
                },
            },
            "Up",
        )
        == [pathlib.Path("reg_2/variation.root"), pathlib.Path("reg_2/new.root")]
    )

    # systematic without override
    assert template_builder._get_ntuple_paths(
        "{SamplePaths}", {}, {"SamplePaths": "path.root"}, {"Name": "variation"}, "Up"
    ) == [pathlib.Path("path.root")]

    caplog.set_level(logging.DEBUG)
    caplog.clear()

    # warning: no region path in template
    assert template_builder._get_ntuple_paths(
        "path.root", {"RegionPath": "region.root"}, {}, {}, ""
    ) == [pathlib.Path("path.root")]
    assert "region override specified, but {RegionPath} not found in default path" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # warning: no region path in template
    assert template_builder._get_ntuple_paths(
        "path.root", {}, {"SamplePaths": "sample.root"}, {}, ""
    ) == [pathlib.Path("path.root")]
    assert "sample override specified, but {SamplePaths} not found in default path" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # error: no override for {RegionPath}
    with pytest.raises(ValueError, match="no path setting found for region region"):
        template_builder._get_ntuple_paths(
            "{RegionPath}", {"Name": "region"}, {}, {}, ""
        )

    # error: no override for {SamplePaths}
    with pytest.raises(ValueError, match="no path setting found for sample sample"):
        template_builder._get_ntuple_paths(
            "{SamplePaths}", {}, {"Name": "sample"}, {}, ""
        )


def test__get_variable():
    assert template_builder._get_variable({"Variable": "jet_pt"}) == "jet_pt"


def test__get_filter():
    # no override
    assert (
        template_builder._get_filter({"Filter": "jet_pt > 0"}, {}, {}, "")
        == "jet_pt > 0"
    )

    # no filter
    assert template_builder._get_filter({}, {}, {}, "") is None

    # systematic with override
    assert (
        template_builder._get_filter(
            {"Filter": "jet_pt > 0"},
            {},
            {"Name": "variation", "Up": {"Filter": "jet_pt > 100"}},
            "Up",
        )
        == "jet_pt > 100"
    )

    # systematic without override
    assert (
        template_builder._get_filter(
            {"Filter": "jet_pt > 0"}, {}, {"Name": "variation"}, "Up"
        )
        == "jet_pt > 0"
    )


def test__get_weight():
    # no override
    assert (
        template_builder._get_weight({}, {"Weight": "weight_mc"}, {}, "") == "weight_mc"
    )

    # no weight
    assert template_builder._get_weight({}, {}, {}, "") is None

    # systematic with override
    assert (
        template_builder._get_weight(
            {},
            {"Weight": "weight_mc"},
            {"Name": "variation", "Up": {"Weight": "weight_modified"}},
            "Up",
        )
        == "weight_modified"
    )

    # systematic without override
    assert (
        template_builder._get_weight(
            {}, {"Weight": "weight_mc"}, {"Name": "variation"}, "Up"
        )
        == "weight_mc"
    )


def test__get_position_in_file():
    # no override
    assert (
        template_builder._get_position_in_file(
            {"Tree": "tree_name"}, {"Name": "Nominal"}, ""
        )
        == "tree_name"
    )

    # systematic with override
    assert (
        template_builder._get_position_in_file(
            {"Tree": "nominal"}, {"Name": "variation", "Up": {"Tree": "up_tree"}}, "Up"
        )
        == "up_tree"
    )

    # systematic without override
    assert (
        template_builder._get_position_in_file(
            {"Tree": "nominal"}, {"Name": "variation"}, "Up"
        )
        == "nominal"
    )


def test__get_binning():
    np.testing.assert_equal(
        template_builder._get_binning({"Binning": [1, 2, 3]}), [1, 2, 3]
    )
    with pytest.raises(NotImplementedError, match="cannot determine binning"):
        template_builder._get_binning({})


def test__Builder():
    builder = template_builder._Builder(pathlib.Path("path"), "file.root", "uproot")
    assert builder.histogram_folder == pathlib.Path("path")
    assert builder.general_path == "file.root"
    assert builder.method == "uproot"


@mock.patch("cabinetry.template_builder._Builder._name_and_save")
@mock.patch("cabinetry.histo.Histogram.from_arrays", return_value="histogram")
@mock.patch(
    "cabinetry.contrib.histogram_creation.from_uproot", return_value=([1], [0.1])
)
def test__Builder_create_histogram(mock_uproot_builder, mock_histo, mock_save):
    # the binning [0] is not a proper binning, but simplifies the comparison
    region = {"Name": "test_region", "Variable": "x", "Binning": [0], "Filter": "x>3"}
    sample = {
        "Name": "sample",
        "Tree": "tree",
        "SamplePaths": "path_to_sample",
        "Weight": "weight_mc",
    }
    systematic = {"Name": "Nominal"}

    builder = template_builder._Builder(pathlib.Path("path"), "{SamplePaths}", "uproot")
    builder._create_histogram(region, sample, systematic, "Nominal")

    # verify the backend call happened properly
    assert mock_uproot_builder.call_args_list == [
        (
            ([pathlib.Path("path_to_sample")], "tree", "x", [0]),
            {"weight": "weight_mc", "selection_filter": "x>3"},
        )
    ]

    # verify the histogram conversion call
    assert mock_histo.call_args_list == [(([0], [1], [0.1]), {})]

    # verify the call for saving
    assert mock_save.call_args_list == [
        (("histogram", region, sample, systematic, "Nominal"), {})
    ]

    # other backends
    builder_unknown = template_builder._Builder(
        pathlib.Path("path"), "{SamplePaths}", "unknown"
    )
    with pytest.raises(NotImplementedError, match="unknown backend unknown"):
        builder_unknown._create_histogram(region, sample, systematic, "Nominal")


@mock.patch("cabinetry.histo.build_name", return_value="name")
def test__Builder__name_and_save(mock_name):
    region = {"Name": "test_region"}
    sample = {"Name": "sample"}
    systematic = {"Name": "Nominal"}

    histogram = mock.MagicMock()

    builder = template_builder._Builder(pathlib.Path("path"), "file.root", "uproot")
    builder._name_and_save(histogram, region, sample, systematic, "Up")

    # check that the naming function was called, the histogram was validated and saved
    assert mock_name.call_args_list == [((region, sample, systematic, "Up"), {})]
    assert histogram.validate.call_args_list == [mock.call("name")]
    assert histogram.save.call_args_list == [mock.call(pathlib.Path("path/name"))]


@mock.patch("cabinetry.template_builder._Builder._name_and_save")
def test__Builder__wrap_custom_template_builder(mock_save):
    histogram = bh.Histogram(bh.axis.Variable([0, 1]))
    region = {"Name": "test_region"}
    sample = {"Name": "sample"}
    systematic = {"Name": "Nominal"}

    def test_func(reg, sam, sys, tem):
        return histogram

    builder = template_builder._Builder(pathlib.Path("path"), "file.root", "uproot")
    wrapped_func = builder._wrap_custom_template_builder(test_func)

    # check the behavior of the wrapped function
    # when called, it should save the returned histogram
    wrapped_func(region, sample, systematic, "Up")

    assert mock_save.call_args_list == [
        ((histogram, region, sample, systematic, "Up"), {})
    ]

    # wrapped function returns wrong type
    def test_func_wrong_return(reg, sam, sys, tem):
        return None

    wrapped_func_wrong_return = builder._wrap_custom_template_builder(
        test_func_wrong_return
    )
    with pytest.raises(TypeError, match="must return a boost_histogram.Histogram"):
        wrapped_func_wrong_return(region, sample, systematic, "Up")


def test_create_histograms():
    config = {"General": {"HistogramFolder": "path/", "InputPath": "file.root"}}
    method = "uproot"

    # no router
    with mock.patch("cabinetry.route.apply_to_all_templates") as mock_apply:
        template_builder.create_histograms(config, method)
        assert len(mock_apply.call_args_list) == 1
        config_call, func_call = mock_apply.call_args_list[0][0]
        assert config_call == config
        assert (
            func_call.__name__ == "_create_histogram"
        )  # could also compare to function
        assert mock_apply.call_args_list[0][1] == {"match_func": None}

    # including a router
    mock_router = mock.MagicMock()
    with mock.patch("cabinetry.route.apply_to_all_templates") as mock_apply:
        template_builder.create_histograms(config, method, router=mock_router)

        # verify wrapper was set
        assert (
            mock_router.template_builder_wrapper.__name__
            == "_wrap_custom_template_builder"
        )

        assert len(mock_apply.call_args_list) == 1
        config_call, func_call = mock_apply.call_args_list[0][0]
        assert config_call == config
        assert func_call.__name__ == "_create_histogram"
        assert mock_apply.call_args_list[0][1] == {
            "match_func": mock_router._find_template_builder_match
        }



project = "cabinetry"
author = "cabinetry developers"
copyright = f"2020, {author}"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_rtd_theme",
    "sphinx_copybutton",
    "sphinx-jsonschema",
    "sphinx_click",
]

import sphinx_rtd_theme

html_theme = "sphinx_rtd_theme"



from collections import namedtuple
import pathlib
from unittest import mock

import numpy as np
import pyhf
import pytest

from cabinetry import fit
from cabinetry import visualize


MockHistogram = namedtuple("MockHistogram", ["bins", "yields", "stdev"])


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (("SR", True), "SR_prefit.pdf"),
        (("SR", False), "SR_postfit.pdf"),
        (("SR 1", True), "SR-1_prefit.pdf"),
        (("SR 1", False), "SR-1_postfit.pdf"),
    ],
)
def test__build_figure_name(test_input, expected):
    assert visualize._build_figure_name(*test_input) == expected


def test__total_yield_uncertainty():
    stdev_list = [np.asarray([0.1, 0.2, 0.1]), np.asarray([0.3, 0.2, 0.1])]
    expected_uncertainties = [0.31622777, 0.28284271, 0.14142136]
    assert np.allclose(
        visualize._total_yield_uncertainty(stdev_list), expected_uncertainties
    )


@mock.patch("cabinetry.visualize._total_yield_uncertainty", return_value=[0.2])
@mock.patch("cabinetry.contrib.matplotlib_visualize.data_MC")
@mock.patch(
    "cabinetry.histo.Histogram.from_config",
    return_value=MockHistogram([0.0, 1.0], [1.0], [0.1]),
)
def test_data_MC_from_histograms(mock_load, mock_draw, mock_stdev):
    """contrib.matplotlib_visualize is only imported depending on the keyword argument,
    so cannot patch via cabinetry.visualize.matplotlib_visualize
    Generally it seems like following the path to the module is preferred, but that
    does not work for the ``data_MC`` case. For some information see also
    https://docs.python.org/3/library/unittest.mock.html#where-to-patch
    """
    config = {
        "General": {"HistogramFolder": "tmp_hist"},
        "Regions": [{"Name": "reg_1", "Variable": "x"}],
        "Samples": [{"Name": "sample_1"}, {"Name": "data", "Data": True}],
    }
    figure_folder = pathlib.Path("tmp")
    histogram_folder = pathlib.Path("tmp_hist")

    visualize.data_MC_from_histograms(
        config, figure_folder=figure_folder, method="matplotlib"
    )

    # the call_args_list contains calls (outer round brackets), first filled with
    # arguments (inner round brackets) and then keyword arguments
    assert mock_load.call_args_list == [
        (
            (
                histogram_folder,
                {"Name": "reg_1", "Variable": "x"},
                {"Name": "data", "Data": True},
                {"Name": "Nominal"},
            ),
            {"modified": True},
        ),
        (
            (
                histogram_folder,
                {"Name": "reg_1", "Variable": "x"},
                {"Name": "sample_1"},
                {"Name": "Nominal"},
            ),
            {"modified": True},
        ),
    ]
    assert mock_stdev.call_args_list == [(([[0.1]],), {})]
    assert mock_draw.call_args_list == [
        (
            (
                [
                    {
                        "label": "data",
                        "isData": True,
                        "yields": [1.0],
                        "variable": "x",
                    },
                    {
                        "label": "sample_1",
                        "isData": False,
                        "yields": [1.0],
                        "variable": "x",
                    },
                ],
                [0.2],
                [0.0, 1.0],
                figure_folder / "reg_1_prefit.pdf",
            ),
            {"log_scale": None},
        )
    ]

    # other plotting method
    with pytest.raises(NotImplementedError, match="unknown backend: unknown"):
        visualize.data_MC_from_histograms(
            config, figure_folder=figure_folder, method="unknown"
        )


@mock.patch("cabinetry.contrib.matplotlib_visualize.data_MC")
@mock.patch("cabinetry.template_builder._get_binning", return_value=np.asarray([1, 2]))
@mock.patch(
    "cabinetry.configuration.get_region_dict",
    return_value={"Name": "region", "Variable": "x"},
)
@mock.patch("cabinetry.tabulate._yields")
@mock.patch("cabinetry.model_utils.calculate_stdev", return_value=[[0.3]])
@mock.patch(
    "cabinetry.model_utils.get_prefit_uncertainties",
    return_value=([0.04956657, 0.0]),
)
@mock.patch(
    "cabinetry.model_utils.get_asimov_parameters",
    return_value=([1.0, 1.0]),
)
def test_data_MC(
    mock_asimov,
    mock_unc,
    mock_stdev,
    mock_table,
    mock_dict,
    mock_bins,
    mock_draw,
    example_spec,
):
    config = {}
    figure_folder = "tmp"
    model_spec = pyhf.Workspace(example_spec).model().spec

    # pre-fit plot
    visualize.data_MC(config, example_spec, figure_folder=figure_folder)

    # Asimov parameter calculation and pre-fit uncertainties
    assert mock_stdev.call_count == 1
    assert mock_asimov.call_args_list[0][0][0].spec == model_spec
    assert mock_unc.call_count == 1
    assert mock_unc.call_args_list[0][0][0].spec == model_spec

    # call to stdev calculation
    assert mock_stdev.call_count == 1
    assert mock_stdev.call_args_list[0][0][0].spec == model_spec
    assert np.allclose(mock_stdev.call_args_list[0][0][1], [1.0, 1.0])
    assert np.allclose(mock_stdev.call_args_list[0][0][2], [0.04956657, 0.0])
    assert np.allclose(
        mock_stdev.call_args_list[0][0][3], np.asarray([[1.0, 0.0], [0.0, 1.0]])
    )
    assert mock_stdev.call_args_list[0][1] == {}

    # yield table
    assert mock_table.call_count == 1
    assert mock_table.call_args_list[0][0][0].spec == model_spec
    assert mock_table.call_args_list[0][0][1] == [np.asarray([[51.839756]])]
    assert mock_table.call_args_list[0][0][2] == [[0.3]]
    assert mock_table.call_args_list[0][0][3] == [np.asarray([475])]
    assert mock_table.call_args_list[0][1] == {}

    assert mock_dict.call_args_list == [[(config, "Signal Region"), {}]]
    assert mock_bins.call_args_list == [[({"Name": "region", "Variable": "x"},), {}]]

    expected_histograms = [
        {
            "label": "Signal",
            "isData": False,
            "yields": np.asarray([51.839756]),
            "variable": "x",
        },
        {
            "label": "Data",
            "isData": True,
            "yields": np.asarray([475]),
            "variable": "x",
        },
    ]
    assert mock_draw.call_count == 1
    assert mock_draw.call_args_list[0][0][0] == expected_histograms
    assert np.allclose(mock_draw.call_args_list[0][0][1], np.asarray([0.3]))
    assert np.allclose(mock_draw.call_args_list[0][0][2], np.asarray([1, 2]))
    assert mock_draw.call_args_list[0][0][3] == pathlib.Path(
        "tmp/Signal-Region_prefit.pdf"
    )
    assert mock_draw.call_args_list[0][1] == {"log_scale": None}

    # post-fit plot and custom scale
    fit_results = fit.FitResults(
        np.asarray([1.01, 1.1]),
        np.asarray([0.03, 0.1]),
        [],
        np.asarray([[1.0, 0.2], [0.2, 1.0]]),
        0.0,
    )
    visualize.data_MC(
        config,
        example_spec,
        figure_folder=figure_folder,
        fit_results=fit_results,
        log_scale=False,
    )

    assert mock_asimov.call_count == 1  # no new call

    # call to stdev calculation
    assert mock_stdev.call_count == 2
    assert mock_stdev.call_args_list[1][0][0].spec == model_spec
    assert np.allclose(mock_stdev.call_args_list[1][0][1], [1.01, 1.1])
    assert np.allclose(mock_stdev.call_args_list[1][0][2], [0.03, 0.1])
    assert np.allclose(
        mock_stdev.call_args_list[1][0][3], np.asarray([[1.0, 0.2], [0.2, 1.0]])
    )
    assert mock_stdev.call_args_list[1][1] == {}

    assert mock_draw.call_count == 2
    # yield at best-fit point is different from pre-fit
    assert np.allclose(mock_draw.call_args_list[1][0][0][0]["yields"], 57.59396892)
    assert np.allclose(mock_draw.call_args_list[1][0][1], np.asarray([0.3]))
    assert np.allclose(mock_draw.call_args_list[1][0][2], np.asarray([1, 2]))
    assert mock_draw.call_args_list[1][0][3] == pathlib.Path(
        "tmp/Signal-Region_postfit.pdf"
    )
    assert mock_draw.call_args_list[1][1] == {"log_scale": False}

    # no yield table
    visualize.data_MC(config, example_spec, include_table=False)
    assert mock_table.call_count == 2  # 2 calls from before

    # unknown plotting method
    with pytest.raises(NotImplementedError, match="unknown backend: unknown"):
        visualize.data_MC(
            config, example_spec, figure_folder=figure_folder, method="unknown"
        )


@mock.patch("cabinetry.contrib.matplotlib_visualize.correlation_matrix")
def test_correlation_matrix(mock_draw):
    corr_mat = np.asarray([[1.0, 0.2, 0.1], [0.2, 1.0, 0.1], [0.1, 0.1, 1.0]])
    corr_mat_pruned = np.asarray([[1.0, 0.2], [0.2, 1.0]])
    labels = ["a", "b", "c"]
    labels_pruned = ["a", "b"]
    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "correlation_matrix.pdf"
    fit_results = fit.FitResults(np.empty(0), np.empty(0), labels, corr_mat, 1.0)

    # pruning with threshold
    visualize.correlation_matrix(
        fit_results, figure_folder=folder_path, pruning_threshold=0.15
    )

    mock_draw.assert_called_once()
    assert np.allclose(mock_draw.call_args[0][0], corr_mat_pruned)
    assert np.any(
        [mock_draw.call_args[0][1][i] == labels[i] for i in range(len(labels_pruned))]
    )
    assert mock_draw.call_args[0][2] == figure_path
    assert mock_draw.call_args[1] == {}

    # pruning of fixed parameter (all zeros in correlation matrix row/column)
    corr_mat_fixed = np.asarray([[1.0, 0.2, 0.0], [0.2, 1.0, 0.0], [0.0, 0.0, 0.0]])
    fit_results_fixed = fit.FitResults(
        np.empty(0), np.empty(0), labels, corr_mat_fixed, 1.0
    )
    visualize.correlation_matrix(fit_results_fixed, figure_folder=folder_path)
    assert np.allclose(mock_draw.call_args_list[1][0][0], corr_mat_pruned)
    assert np.any(
        [
            mock_draw.call_args_list[1][0][1][i] == labels[i]
            for i in range(len(labels_pruned))
        ]
    )

    # unknown plotting method
    with pytest.raises(NotImplementedError, match="unknown backend: unknown"):
        visualize.correlation_matrix(
            fit_results, figure_folder=folder_path, method="unknown"
        )


@mock.patch("cabinetry.contrib.matplotlib_visualize.pulls")
def test_pulls(mock_draw):
    bestfit = np.asarray([0.8, 1.0, 1.05, 1.1])
    uncertainty = np.asarray([0.9, 1.0, 0.03, 0.7])
    labels = ["a", "b", "staterror_region[bin_0]", "c"]
    exclude_list = ["a"]
    folder_path = "tmp"
    fit_results = fit.FitResults(bestfit, uncertainty, labels, np.empty(0), 1.0)

    filtered_bestfit = np.asarray([1.0, 1.1])
    filtered_uncertainty = np.asarray([1.0, 0.7])
    filtered_labels = np.asarray(["b", "c"])
    figure_path = pathlib.Path(folder_path) / "pulls.pdf"

    # with filtering
    visualize.pulls(
        fit_results,
        figure_folder=folder_path,
        exclude_list=exclude_list,
        method="matplotlib",
    )

    mock_draw.assert_called_once()
    assert np.allclose(mock_draw.call_args[0][0], filtered_bestfit)
    assert np.allclose(mock_draw.call_args[0][1], filtered_uncertainty)
    assert np.any(
        [
            mock_draw.call_args[0][2][i] == filtered_labels[i]
            for i in range(len(filtered_labels))
        ]
    )
    assert mock_draw.call_args[0][3] == figure_path
    assert mock_draw.call_args[1] == {}

    # without filtering via list, but with staterror removal
    # and fixed parameter removal
    fit_results.uncertainty[0] = 0.0

    bestfit_expected = np.asarray([1.0, 1.1])
    uncertainty_expected = np.asarray([1.0, 0.7])
    labels_expected = ["b", "c"]
    visualize.pulls(fit_results, figure_folder=folder_path, method="matplotlib")

    assert np.allclose(mock_draw.call_args[0][0], bestfit_expected)
    assert np.allclose(mock_draw.call_args[0][1], uncertainty_expected)
    assert np.any(
        [
            mock_draw.call_args[0][2][i] == labels_expected[i]
            for i in range(len(labels_expected))
        ]
    )
    assert mock_draw.call_args[0][3] == figure_path
    assert mock_draw.call_args[1] == {}

    # unknown plotting method
    with pytest.raises(NotImplementedError, match="unknown backend: unknown"):
        visualize.pulls(
            fit_results,
            figure_folder=folder_path,
            exclude_list=exclude_list,
            method="unknown",
        )


@mock.patch("cabinetry.contrib.matplotlib_visualize.ranking")
def test_ranking(mock_draw):
    bestfit = np.asarray([1.2, 0.1])
    uncertainty = np.asarray([0.2, 0.8])
    labels = ["staterror_a", "modeling"]
    impact_prefit_up = np.asarray([0.1, 0.5])
    impact_prefit_down = np.asarray([-0.2, -0.4])
    impact_postfit_up = np.asarray([0.1, 0.4])
    impact_postfit_down = np.asarray([-0.2, -0.3])
    ranking_results = fit.RankingResults(
        bestfit,
        uncertainty,
        labels,
        impact_prefit_up,
        impact_prefit_down,
        impact_postfit_up,
        impact_postfit_down,
    )
    folder_path = "tmp"

    figure_path = pathlib.Path(folder_path) / "ranking.pdf"
    bestfit_expected = np.asarray([0.1, 1.2])
    uncertainty_expected = np.asarray([0.8, 0.2])
    labels_expected = ["modeling", "staterror_a"]

    visualize.ranking(ranking_results, figure_folder=folder_path)
    assert mock_draw.call_count == 1
    assert np.allclose(mock_draw.call_args[0][0], bestfit_expected)
    assert np.allclose(mock_draw.call_args[0][1], uncertainty_expected)
    for i_lab, label in enumerate(mock_draw.call_args[0][2]):
        assert label == labels_expected[i_lab]
    assert np.allclose(mock_draw.call_args[0][3], impact_prefit_up[::-1])
    assert np.allclose(mock_draw.call_args[0][4], impact_prefit_down[::-1])
    assert np.allclose(mock_draw.call_args[0][5], impact_postfit_up[::-1])
    assert np.allclose(mock_draw.call_args[0][6], impact_postfit_down[::-1])
    assert mock_draw.call_args[0][7] == figure_path
    assert mock_draw.call_args[1] == {}

    # maximum parameter amount specified
    visualize.ranking(ranking_results, figure_folder=folder_path, max_pars=1)
    assert mock_draw.call_count == 2
    assert np.allclose(mock_draw.call_args[0][0], bestfit_expected[0])
    assert np.allclose(mock_draw.call_args[0][1], uncertainty_expected[0])
    assert mock_draw.call_args[0][2] == labels_expected[0]
    assert np.allclose(mock_draw.call_args[0][3], impact_prefit_up[1])
    assert np.allclose(mock_draw.call_args[0][4], impact_prefit_down[1])
    assert np.allclose(mock_draw.call_args[0][5], impact_postfit_up[1])
    assert np.allclose(mock_draw.call_args[0][6], impact_postfit_down[1])
    assert mock_draw.call_args[0][7] == figure_path
    assert mock_draw.call_args[1] == {}

    # unknown plotting method
    with pytest.raises(NotImplementedError, match="unknown backend: unknown"):
        visualize.ranking(ranking_results, figure_folder=folder_path, method="unknown")


@mock.patch(
    "cabinetry.histo.Histogram.from_path",
    side_effect=[
        MockHistogram([0.0, 1.0], [2.0], [0.2]),
        MockHistogram([0.0, 1.0], [3.0], [0.3]),
    ]
    * 3,
)
@mock.patch(
    "cabinetry.histo.Histogram.from_config",
    return_value=MockHistogram([0.0, 1.0], [1.0], [0.1]),
)
@mock.patch("cabinetry.contrib.matplotlib_visualize.templates")
def test_templates(mock_draw, mock_histo_config, mock_histo_path, tmp_path):
    # the side effects are repeated for the patched Histogram.from_path
    # to check all relevant behavior (including the unknown backend check)
    nominal_path = tmp_path / "region_sample_Nominal_modified.npz"
    up_path = tmp_path / "region_sample_sys_Up_modified.npz"
    down_path = tmp_path / "region_sample_sys_Down_modified.npz"
    region = {"Name": "region", "Variable": "x"}
    sample = {"Name": "sample"}
    config = {
        "General": {"HistogramFolder": tmp_path},
        "Regions": [region],
        "Samples": [sample, {"Name": "data", "Data": True}],
        "Systematics": [{"Name": "sys"}],
    }

    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "templates/region_sample_sys.pdf"

    # add fake histograms for glob
    nominal_path.touch()
    up_path.touch()
    down_path.touch()

    # also add a file that matches pattern but is not needed
    (tmp_path / "region_sample_sys_unknown_modified.npz").touch()

    visualize.templates(config, figure_folder=folder_path)

    assert mock_histo_config.call_args_list == [
        [(tmp_path, region, sample, {"Name": "Nominal"}), {}]
    ]
    assert mock_histo_path.call_args_list == [[(down_path,), {}], [(up_path,), {}]]

    nominal = {"yields": [1.0], "stdev": [0.1]}
    up = {"yields": [3.0], "stdev": [0.3]}
    down = {"yields": [2.0], "stdev": [0.2]}
    bins = [0.0, 1.0]
    assert mock_draw.call_args_list == [
        [(nominal, up, down, bins, "x", figure_path), {}]
    ]

    # unknown plotting method
    with pytest.raises(NotImplementedError, match="unknown backend: unknown"):
        visualize.templates(config, figure_folder=folder_path, method="unknown")

    # remove files for variation histograms
    up_path.unlink()
    down_path.unlink()

    visualize.templates(config, figure_folder=folder_path)
    assert mock_draw.call_count == 1  # no new call, since no variations found


@mock.patch("cabinetry.contrib.matplotlib_visualize.scan")
def test_scan(mock_draw):
    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "scan_a_0.pdf"

    par_name = "a[0]"
    par_mle = 1.5
    par_unc = 0.2
    par_vals = np.asarray([1.3, 1.5, 1.7])
    par_nlls = np.asarray([0.9, 0.0, 1.1])
    scan_results = fit.ScanResults(par_name, par_mle, par_unc, par_vals, par_nlls)

    visualize.scan(scan_results, figure_folder=folder_path)

    assert mock_draw.call_count == 1
    assert mock_draw.call_args[0][0] == par_name
    assert mock_draw.call_args[0][1] == par_mle
    assert mock_draw.call_args[0][2] == par_unc
    assert np.allclose(mock_draw.call_args[0][3], par_vals)
    assert np.allclose(mock_draw.call_args[0][4], par_nlls)
    assert mock_draw.call_args[0][5] == figure_path
    assert mock_draw.call_args[1] == {}

    # unknown plotting method
    with pytest.raises(NotImplementedError, match="unknown backend: unknown"):
        visualize.scan(scan_results, figure_folder=folder_path, method="unknown")


@mock.patch("cabinetry.contrib.matplotlib_visualize.limit")
def test_limit(mock_draw):
    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "limit.pdf"

    observed_CLs = np.asarray([0.75, 0.32, 0.02])
    expected_CLs = np.asarray([[0.1, 0.2, 0.3, 0.4, 0.5] for _ in range(3)])
    poi_values = np.asarray([0, 1, 2])
    limit_results = fit.LimitResults(
        3.0, np.empty(5), observed_CLs, expected_CLs, poi_values
    )

    visualize.limit(limit_results, figure_folder=folder_path)

    assert mock_draw.call_count == 1
    assert np.allclose(mock_draw.call_args[0][0], limit_results.observed_CLs)
    assert np.allclose(mock_draw.call_args[0][1], limit_results.expected_CLs)
    assert np.allclose(mock_draw.call_args[0][2], limit_results.poi_values)
    assert mock_draw.call_args[0][3] == figure_path
    assert mock_draw.call_args[1] == {}

    # unknown plotting method
    with pytest.raises(NotImplementedError, match="unknown backend: unknown"):
        visualize.limit(limit_results, figure_folder=folder_path, method="unknown")



import pytest
import uproot3 as uproot


class Utils:
    @staticmethod
    def create_ntuple(fname, treename, varname, var_array, weightname, weight_array):
        with uproot.recreate(fname) as f:
            f[treename] = uproot.newtree({varname: "float64", weightname: "float64"})
            f[treename].extend({varname: var_array, weightname: weight_array})


@pytest.fixture
def utils():
    return Utils


@pytest.fixture
def example_spec():
    spec = {
        "channels": [
            {
                "name": "Signal Region",
                "samples": [
                    {
                        "data": [51.839756],
                        "modifiers": [
                            {
                                "data": [2.5695188],
                                "name": "staterror_Signal-Region",
                                "type": "staterror",
                            },
                            {
                                "data": None,
                                "name": "Signal strength",
                                "type": "normfactor",
                            },
                        ],
                        "name": "Signal",
                    }
                ],
            }
        ],
        "measurements": [
            {
                "config": {
                    "parameters": [
                        {
                            "name": "staterror_Signal-Region",
                            "fixed": True,
                            "inits": [1.1],
                        }
                    ],
                    "poi": "Signal strength",
                },
                "name": "My fit",
            }
        ],
        "observations": [{"data": [475], "name": "Signal Region"}],
        "version": "1.0.0",
    }
    return spec


@pytest.fixture
def example_spec_multibin():
    spec = {
        "channels": [
            {
                "name": "region_1",
                "samples": [
                    {
                        "data": [25, 5],
                        "modifiers": [
                            {
                                "data": [5, 2],
                                "name": "staterror_region_1",
                                "type": "staterror",
                            },
                            {
                                "data": None,
                                "name": "Signal strength",
                                "type": "normfactor",
                            },
                        ],
                        "name": "Signal",
                    }
                ],
            },
            {
                "name": "region_2",
                "samples": [
                    {
                        "data": [8],
                        "modifiers": [
                            {
                                "data": [1],
                                "name": "staterror_region_2",
                                "type": "staterror",
                            },
                            {
                                "data": None,
                                "name": "Signal strength",
                                "type": "normfactor",
                            },
                        ],
                        "name": "Signal",
                    }
                ],
            },
        ],
        "measurements": [
            {"config": {"parameters": [], "poi": "Signal strength"}, "name": "My fit"}
        ],
        "observations": [
            {"data": [35, 8], "name": "region_1"},
            {"data": [10], "name": "region_2"},
        ],
        "version": "1.0.0",
    }
    return spec


@pytest.fixture
def example_spec_shapefactor():
    spec = {
        "channels": [
            {
                "name": "Signal Region",
                "samples": [
                    {
                        "data": [20, 10],
                        "modifiers": [
                            {
                                "data": None,
                                "name": "shape factor",
                                "type": "shapefactor",
                            },
                            {
                                "data": None,
                                "name": "Signal strength",
                                "type": "normfactor",
                            },
                        ],
                        "name": "Signal",
                    }
                ],
            }
        ],
        "measurements": [
            {
                "config": {"parameters": [], "poi": "Signal strength"},
                "name": "shapefactor fit",
            }
        ],
        "observations": [{"data": [25, 8], "name": "Signal Region"}],
        "version": "1.0.0",
    }
    return spec


@pytest.fixture
def example_spec_with_background():
    spec = {
        "channels": [
            {
                "name": "Signal Region",
                "samples": [
                    {
                        "data": [50],
                        "modifiers": [
                            {
                                "data": [5],
                                "name": "staterror_Signal-Region",
                                "type": "staterror",
                            },
                            {
                                "data": None,
                                "name": "Signal strength",
                                "type": "normfactor",
                            },
                        ],
                        "name": "Signal",
                    },
                    {
                        "data": [150],
                        "modifiers": [
                            {
                                "data": [7],
                                "name": "staterror_Signal-Region",
                                "type": "staterror",
                            }
                        ],
                        "name": "Background",
                    },
                ],
            }
        ],
        "measurements": [
            {
                "config": {
                    "parameters": [
                        {
                            "name": "Signal strength",
                            "bounds": [[0, 10]],
                            "inits": [1.0],
                        }
                    ],
                    "poi": "Signal strength",
                },
                "name": "signal plus background",
            }
        ],
        "observations": [{"data": [160], "name": "Signal Region"}],
        "version": "1.0.0",
    }
    return spec



import functools
import logging
import pathlib
from typing import Any, Dict, List, Optional, Union

import boost_histogram as bh
import numpy as np

from . import configuration
from . import histo
from . import route


log = logging.getLogger(__name__)


def _check_for_override(
    systematic: Dict[str, Any], template: str, option: str
) -> Optional[Union[str, List[str]]]:
    """Returns an override if specified by a template of a systematic.

    Given a systematic and a string specifying which template is currently under
    consideration, check whether the systematic defines an override for an option.
    Return the override if it exists, otherwise return None.

    Args:
        systematic (Dict[str, Any]): containing all systematic information
        template (str): template to consider: "Nominal", "Up", "Down"
        option (str): the option for which the presence of an override is checked

    Returns:
        Optional[Union[str, List[str]]]: either None if no override exists, or the
        override
    """
    return systematic.get(template, {}).get(option, None)


def _get_ntuple_paths(
    general_path: str,
    region: Dict[str, Any],
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    template: str,
) -> List[pathlib.Path]:
    """Returns the paths to ntuples for a region-sample-systematic-template.

    A path is built starting from the path specified in the general options in the
    configuration file. This path can contain placeholders for region- and sample-
    specific overrides, via ``{Region}`` and ``{Sample}``. For non-nominal templates, it
    is possible to override the sample path if the ``SamplePaths`` option is specified
    for the template. If ``SamplePaths`` is a list, return a list of paths (one per
    entry in the list).

    Args:
        general_path (str): path specified in general settings, with sections that can
            be overridden by region / sample settings
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (str): which template is considered: "Nominal", "Up", "Down"

    Returns:
        List[pathlib.Path]: list of paths to ntuples
    """
    # obtain region and sample paths, if they are defined
    region_path = region.get("RegionPath", None)
    sample_paths = sample.get("SamplePaths", None)

    # check whether a systematic is being processed, and whether overrides exist
    if systematic.get("Name", "Nominal") != "Nominal":
        # determine whether the template has an override for RegionPath specified
        region_override = _check_for_override(systematic, template, "RegionPath")
        if region_override is not None:
            region_path = region_override

        # check for SamplePaths override
        sample_override = _check_for_override(systematic, template, "SamplePaths")
        if sample_override is not None:
            sample_paths = sample_override

    region_template_exists = "{RegionPath}" in general_path
    if region_path is not None:
        if not region_template_exists:
            log.warning(
                "region override specified, but {RegionPath} not found in default path"
            )
        general_path = general_path.replace("{RegionPath}", region_path)
    elif region_template_exists:
        raise ValueError(f"no path setting found for region {region['Name']}")

    sample_template_exists = "{SamplePaths}" in general_path
    if sample_paths is not None:
        if not sample_template_exists:
            log.warning(
                "sample override specified, but {SamplePaths} not found in default path"
            )
        # SamplePaths can be a list, so need to construct all possible paths
        sample_paths = configuration._convert_setting_to_list(sample_paths)
        path_list = []
        for sample_path in sample_paths:
            path_list.append(general_path.replace("{SamplePaths}", sample_path))
    elif sample_template_exists:
        raise ValueError(f"no path setting found for sample {sample['Name']}")
    else:
        # no need for multiple paths, and no SamplePaths are present, so turn
        # the existing path into a list
        path_list = [general_path]

    # convert the contents of path_lists to paths and return them
    paths = [pathlib.Path(path) for path in path_list]
    return paths


def _get_variable(region: Dict[str, Any]) -> str:
    """Returns the variable the histogram will be binned in.

    Args:
        region (Dict[str, Any]): containing all region information

    Returns:
        str: name of variable to bin histogram in
    """
    axis_variable = region["Variable"]
    return axis_variable


def _get_filter(
    region: Dict[str, Any],
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    template: str,
) -> Optional[str]:
    """Returns the filter to be applied for event selection.

    For non-nominal templates, overrides the nominal filter if an alternative is
    specified for the template.

    Args:
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (str): which template is considered: "Nominal", "Up", "Down"

    Returns:
        Optional[str]: expression for the filter to be used, or None for no filtering
    """
    selection_filter = region.get("Filter", None)
    # check whether a systematic is being processed
    if systematic.get("Name", "Nominal") != "Nominal":
        # determine whether the template has an override specified
        selection_filter_override = _check_for_override(systematic, template, "Filter")
        if selection_filter_override is not None:
            selection_filter = selection_filter_override
    return selection_filter


def _get_weight(
    region: Dict[str, Any],
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    template: str,
) -> Optional[str]:
    """Returns the weight to be used for events in histograms.

    For non-nominal templates, overrides the nominal weight if an alternative is
    specified for the template.

    Args:
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (str): which template is considered: "Nominal", "Up", "Down"

    Returns:
        Optional[str]: weight used for events when filled into histograms, or None for
        no weight
    """
    weight = sample.get("Weight", None)
    # check whether a systematic is being processed
    if systematic.get("Name", "Nominal") != "Nominal":
        # determine whether the template has an override specified
        weight_override = _check_for_override(systematic, template, "Weight")
        if weight_override is not None:
            weight = weight_override
    return weight


def _get_position_in_file(
    sample: Dict[str, Any], systematic: Dict[str, Any], template: str
) -> str:
    """Returns the location of data within a file (e.g. a tree name).

    For non-nominal templates, overrides the nominal position if an alternative is
    specified for the template.

    Args:
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (str): which template is considered: "Nominal", "Up", "Down"

    Returns:
        str: where in the file to find the data (the name of a tree)
    """
    position = sample["Tree"]
    # check whether a systematic is being processed
    if systematic.get("Name", "Nominal") != "Nominal":
        # determine whether the template has an override specified
        position_override = _check_for_override(systematic, template, "Tree")
        if position_override is not None:
            position = position_override
    return position


def _get_binning(region: Dict[str, Any]) -> np.ndarray:
    """Returns the binning to be used in a region.

    Should eventually also support other ways of specifying bins, such as the amount of
    bins and the range to bin in.

    Args:
        region (Dict[str, Any]): containing all region information

    Raises:
        NotImplementedError: when the binning is not explicitly defined

    Returns:
        numpy.ndarray: bin boundaries to be used for histogram
    """
    if not region.get("Binning", False):
        raise NotImplementedError("cannot determine binning")

    return np.asarray(region["Binning"])


class _Builder:
    """Handles the instructions for backends to create histograms."""

    def __init__(
        self, histogram_folder: pathlib.Path, general_path: str, method: str
    ) -> None:
        """Creates an instance, sets histogram folder, path template and method.

        Args:
            histogram_folder (pathlib.Path): folder to save the histograms to
            general_path (str): template for paths to input files for histogram building
            method (str): backend to use for histogram production
        """
        self.histogram_folder = histogram_folder
        self.general_path = general_path
        self.method = method

    def _create_histogram(
        self,
        region: Dict[str, Any],
        sample: Dict[str, Any],
        systematic: Dict[str, Any],
        template: str,
    ) -> None:
        """Creates a histogram and writes it to a file.

        The histogram is created for the region-sample-systematic-template specified in
        the argument.

        Args:
            region (Dict[str, Any]): specifying region information
            sample (Dict[str, Any]): specifying sample information
            systematic (Dict[str, Any]): specifying systematic information
            template (str): name of the template: "Nominal", "Up", "Down"

        Raises:
            NotImplementedError: when requesting an unknown backend
        """
        ntuple_paths = _get_ntuple_paths(
            self.general_path, region, sample, systematic, template
        )
        pos_in_file = _get_position_in_file(sample, systematic, template)
        variable = _get_variable(region)
        bins = _get_binning(region)
        weight = _get_weight(region, sample, systematic, template)
        selection_filter = _get_filter(region, sample, systematic, template)

        # obtain the histogram
        if self.method == "uproot":
            from .contrib import histogram_creation

            yields, stdev = histogram_creation.from_uproot(
                ntuple_paths,
                pos_in_file,
                variable,
                bins,
                weight=weight,
                selection_filter=selection_filter,
            )

        else:
            raise NotImplementedError(f"unknown backend {self.method}")

        # store information in a Histogram instance and save it
        histogram = histo.Histogram.from_arrays(bins, yields, stdev)
        self._name_and_save(histogram, region, sample, systematic, template)

    def _name_and_save(
        self,
        histogram: histo.Histogram,
        region: Dict[str, Any],
        sample: Dict[str, Any],
        systematic: Dict[str, Any],
        template: str,
    ) -> None:
        """Generates a unique name for a histogram and saves the histogram.

        Args:
            histogram (histo.Histogram): histogram to save
            region (Dict[str, Any]): dict with region information
            sample (Dict[str, Any]): dict with sample information
            systematic (Dict[str, Any]): dict with systematic information
            template (str): name of the template: "Nominal", "Up", "Down"
        """
        # generate a name for the histogram
        histogram_name = histo.build_name(region, sample, systematic, template)

        # check the histogram for common issues
        histogram.validate(histogram_name)

        # save it
        histo_path = self.histogram_folder / histogram_name
        histogram.save(histo_path)

    def _wrap_custom_template_builder(
        self, func: route.UserTemplateFunc
    ) -> route.ProcessorFunc:
        """Returns function that executes custom template builder and saves histogram.

        Wrapper for custom template builder functions that return a
        ``boost_histogram.Histogram``.

        Args:
            func (cabinetry.route.UserTemplateFunc): user-defined template builder

        Returns:
            cabinetry.route.ProcessorFunc: wrapped template builder
        """
        # decorating this with functools.wraps will keep the name of the wrapped
        # function the same, however the signature of the wrapped function is slightly
        # different (the return value becomes None)
        @functools.wraps(func)
        def wrapper(
            region: Dict[str, Any],
            sample: Dict[str, Any],
            systematic: Dict[str, Any],
            template: str,
        ) -> None:
            """Executes a user-defined function that returns a histogram and saves it.

            Returns None, to turn the user-defined function into a ProcessorFunc when
            wrapped with this.

            Args:
                region (Dict[str, Any]): dict with region information
                sample (Dict[str, Any]): dict with sample information
                systematic (Dict[str, Any]): dict with systematic information
                template (str): name of the template: "Nominal", "Up", "Down"
            """
            histogram = func(region, sample, systematic, template)
            if not isinstance(histogram, bh.Histogram):
                raise TypeError(
                    f"{func.__name__} must return a boost_histogram.Histogram"
                )
            self._name_and_save(
                histo.Histogram(histogram), region, sample, systematic, template
            )

        return wrapper


def create_histograms(
    config: Dict[str, Any],
    method: str = "uproot",
    router: Optional[route.Router] = None,
) -> None:
    """Produces all required histograms specified by the configuration file.

    Uses either a default method specified via ``method``, or a custom user-defined
    override through ``router``.

    Args:
        config (Dict[str, Any]): cabinetry configuration
        method (str, optional): backend to use for histogram production, defaults to
            "uproot"
        router (Optional[route.Router], optional): instance of cabinetry.route.Router
            that contains user-defined overrides, defaults to None
    """
    # create an instance of the class doing the template building
    histogram_folder = pathlib.Path(config["General"]["HistogramFolder"])
    general_path = config["General"]["InputPath"]
    builder = _Builder(histogram_folder, general_path, method)

    match_func: Optional[route.MatchFunc] = None
    if router is not None:
        # specify the wrapper for user-defined functions
        router.template_builder_wrapper = builder._wrap_custom_template_builder
        # get a function that can be queried to return a user-defined template builder
        match_func = router._find_template_builder_match

    route.apply_to_all_templates(
        config, builder._create_histogram, match_func=match_func
    )



import numpy as np
import pyhf
import pytest

from cabinetry import tabulate


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (("abc", 0), "abc\nbin 1"),
        (("abc", 2), "\nbin 3"),
    ],
)
def test__header_name(test_input, expected):
    assert tabulate._header_name(*test_input) == expected


def test__yields(example_spec_multibin, example_spec_with_background):
    # multiple channels
    model = pyhf.Workspace(example_spec_multibin).model()
    yields = [np.asarray([[25.0, 5.0]]), np.asarray([[8.0]])]
    total_stdev = [[5.0, 2.0], [1.0]]
    data = [np.asarray([35, 8]), np.asarray([10])]

    yield_table = tabulate._yields(model, yields, total_stdev, data)
    assert yield_table == [
        {
            "sample": "Signal",
            "region_1\nbin 1": "25.00",
            "\nbin 2": "5.00",
            "region_2\nbin 1": "8.00",
        },
        {
            "sample": "total",
            "region_1\nbin 1": "25.00 \u00B1 5.00",
            "\nbin 2": "5.00 \u00B1 2.00",
            "region_2\nbin 1": "8.00 \u00B1 1.00",
        },
        {
            "sample": "data",
            "region_1\nbin 1": "35.00",
            "\nbin 2": "8.00",
            "region_2\nbin 1": "10.00",
        },
    ]

    # multiple samples
    model = pyhf.Workspace(example_spec_with_background).model()
    yields = [np.asarray([[150.0], [50.0]])]
    total_stdev = [[8.60]]
    data = [np.asarray([160])]

    yield_table = tabulate._yields(model, yields, total_stdev, data)
    assert yield_table == [
        {"sample": "Background", "Signal Region\nbin 1": "150.00"},
        {"sample": "Signal", "Signal Region\nbin 1": "50.00"},
        {"sample": "total", "Signal Region\nbin 1": "200.00 \u00B1 8.60"},
        {"sample": "data", "Signal Region\nbin 1": "160.00"},
    ]



import logging
from unittest import mock

import iminuit
import numpy as np
import pytest

from cabinetry import fit
from cabinetry import model_utils


def test_FitResults():
    bestfit = np.asarray([1.0])
    uncertainty = np.asarray([0.1])
    labels = ["par_a"]
    corr_mat = np.asarray([[1.0]])
    best_twice_nll = 2.0
    fit_results = fit.FitResults(bestfit, uncertainty, labels, corr_mat, best_twice_nll)
    assert np.allclose(fit_results.bestfit, bestfit)
    assert np.allclose(fit_results.uncertainty, uncertainty)
    assert fit_results.labels == labels
    assert np.allclose(fit_results.corr_mat, corr_mat)
    assert fit_results.best_twice_nll == best_twice_nll
    assert fit_results.goodness_of_fit == -1


def test_RankingResults():
    bestfit = np.asarray([1.0])
    uncertainty = np.asarray([0.1])
    labels = ["par_a"]
    prefit_up = np.asarray([0.3])
    prefit_down = np.asarray([-0.3])
    postfit_up = np.asarray([0.2])
    postfit_down = np.asarray([-0.2])
    ranking_results = fit.RankingResults(
        bestfit, uncertainty, labels, prefit_up, prefit_down, postfit_up, postfit_down
    )
    assert np.allclose(ranking_results.bestfit, bestfit)
    assert np.allclose(ranking_results.uncertainty, uncertainty)
    assert ranking_results.labels == labels
    assert np.allclose(ranking_results.prefit_up, prefit_up)
    assert np.allclose(ranking_results.prefit_down, prefit_down)
    assert np.allclose(ranking_results.postfit_up, postfit_up)
    assert np.allclose(ranking_results.postfit_down, postfit_down)


def test_ScanResults():
    name = "par_a"
    bestfit = 1.2
    uncertainty = 0.3
    parameter_values = np.asarray([0.9, 1.2, 1.5])
    delta_nlls = np.asarray([1.0, 0.0, 1.0])
    scan_results = fit.ScanResults(
        name, bestfit, uncertainty, parameter_values, delta_nlls
    )
    assert scan_results.name == name
    assert scan_results.bestfit == bestfit
    assert scan_results.uncertainty == uncertainty
    assert np.allclose(scan_results.parameter_values, parameter_values)
    assert np.allclose(scan_results.delta_nlls, delta_nlls)


def test_LimitResults():
    observed_limit = 3.0
    expected_limit = np.asarray([1.0, 2.0, 3.0, 4.0, 5.0])
    observed_CLs = np.asarray([0.05])
    expected_CLs = np.asarray([0.01, 0.02, 0.05, 0.07, 0.10])
    poi_values = np.asarray([3.0])
    limit_results = fit.LimitResults(
        observed_limit, expected_limit, observed_CLs, expected_CLs, poi_values
    )
    assert limit_results.observed_limit == observed_limit
    assert np.allclose(limit_results.expected_limit, expected_limit)
    assert np.allclose(limit_results.observed_CLs, observed_CLs)
    assert np.allclose(limit_results.expected_CLs, expected_CLs)
    assert np.allclose(limit_results.poi_values, poi_values)


def test_print_results(caplog):
    caplog.set_level(logging.DEBUG)

    bestfit = np.asarray([1.0, 2.0])
    uncertainty = np.asarray([0.1, 0.3])
    labels = ["param_A", "param_B"]
    fit_results = fit.FitResults(bestfit, uncertainty, labels, np.empty(0), 0.0)

    fit.print_results(fit_results)
    assert "param_A:  1.0000 +/- 0.1000" in [rec.message for rec in caplog.records]
    assert "param_B:  2.0000 +/- 0.3000" in [rec.message for rec in caplog.records]
    caplog.clear()


# skip a "RuntimeWarning: numpy.ufunc size changed" warning
# due to different numpy versions used in dependencies
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@mock.patch("cabinetry.fit._run_minos")
def test__fit_model_pyhf(mock_minos, example_spec, example_spec_multibin):
    model, data = model_utils.model_and_data(example_spec)
    fit_results = fit._fit_model_pyhf(model, data)
    assert np.allclose(fit_results.bestfit, [1.1, 8.32984849])
    assert np.allclose(fit_results.uncertainty, [0.0, 0.38099445])
    assert fit_results.labels == ["staterror_Signal-Region", "Signal strength"]
    assert np.allclose(fit_results.best_twice_nll, 7.90080379)
    assert np.allclose(fit_results.corr_mat, [[0.0, 0.0], [0.0, 1.0]])

    # Asimov fit, with fixed gamma (fixed not to Asimov MLE)
    model, data = model_utils.model_and_data(example_spec, asimov=True)
    fit_results = fit._fit_model_pyhf(model, data)
    # the gamma factor is multiplicative and fixed to 1.1, so the
    # signal strength needs to be 1/1.1 to compensate
    assert np.allclose(fit_results.bestfit, [1.1, 0.90917877])
    assert np.allclose(fit_results.uncertainty, [0.0, 0.12623179])
    assert fit_results.labels == ["staterror_Signal-Region", "Signal strength"]
    assert np.allclose(fit_results.best_twice_nll, 5.68851093)
    assert np.allclose(fit_results.corr_mat, [[0.0, 0.0], [0.0, 1.0]])

    # parameters held constant via keyword argument
    model, data = model_utils.model_and_data(example_spec_multibin)
    init_pars = model.config.suggested_init()
    init_pars[0] = 0.9
    init_pars[1] = 1.1
    fix_pars = model.config.suggested_fixed()
    fix_pars[0] = True
    fix_pars[1] = True
    fit_results = fit._fit_model_pyhf(
        model, data, init_pars=init_pars, fix_pars=fix_pars
    )
    assert np.allclose(fit_results.bestfit, [0.9, 1.1, 1.48041923, 0.97511112])
    assert np.allclose(fit_results.uncertainty, [0.0, 0.0, 0.20694465, 0.11792837])
    assert np.allclose(fit_results.best_twice_nll, 10.4531891)

    # including minos, one parameter is unknown
    model, data = model_utils.model_and_data(example_spec)
    fit_results = fit._fit_model_pyhf(model, data, minos=["Signal strength", "abc"])
    assert mock_minos.call_count == 1
    # first argument to minos call is the Minuit instance
    assert mock_minos.call_args[0][1] == ["x1"]
    assert mock_minos.call_args[0][2] == ["staterror_Signal-Region", "Signal strength"]
    assert mock_minos.call_args[1] == {}


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@mock.patch("cabinetry.fit._run_minos")
def test__fit_model_custom(mock_minos, example_spec, example_spec_multibin):
    model, data = model_utils.model_and_data(example_spec)
    fit_results = fit._fit_model_custom(model, data)
    assert np.allclose(fit_results.bestfit, [1.1, 8.32985794])
    assert np.allclose(fit_results.uncertainty, [0.0, 0.38153392])
    assert fit_results.labels == ["staterror_Signal-Region", "Signal strength"]
    assert np.allclose(fit_results.best_twice_nll, 7.90080378)
    assert np.allclose(fit_results.corr_mat, [[0.0, 0.0], [0.0, 1.0]])

    # Asimov fit, with fixed gamma (fixed not to Asimov MLE)
    model, data = model_utils.model_and_data(example_spec, asimov=True)
    fit_results = fit._fit_model_custom(model, data)
    # the gamma factor is multiplicative and fixed to 1.1, so the
    # signal strength needs to be 1/1.1 to compensate
    assert np.allclose(fit_results.bestfit, [1.1, 0.90917877])
    assert np.allclose(fit_results.uncertainty, [0.0, 0.12623172])
    assert fit_results.labels == ["staterror_Signal-Region", "Signal strength"]
    assert np.allclose(fit_results.best_twice_nll, 5.68851093)
    assert np.allclose(fit_results.corr_mat, [[0.0, 0.0], [0.0, 1.0]])

    # parameters held constant via keyword argument
    model, data = model_utils.model_and_data(example_spec_multibin)
    init_pars = model.config.suggested_init()
    init_pars[0] = 0.9
    init_pars[1] = 1.1
    fix_pars = model.config.suggested_fixed()
    fix_pars[0] = True
    fix_pars[1] = True
    fit_results = fit._fit_model_custom(
        model, data, init_pars=init_pars, fix_pars=fix_pars
    )
    assert np.allclose(fit_results.bestfit, [0.9, 1.1, 1.48041923, 0.97511112])
    assert np.allclose(fit_results.uncertainty, [0.0, 0.0, 0.20694409, 0.11792805])
    assert np.allclose(fit_results.best_twice_nll, 10.45318909)

    # including minos
    model, data = model_utils.model_and_data(example_spec)
    fit_results = fit._fit_model_custom(model, data, minos=["Signal strength"])
    assert mock_minos.call_count == 1
    # first argument to minos call is the Minuit instance
    assert mock_minos.call_args[0][1] == ["Signal strength"]
    assert mock_minos.call_args[0][2] == ["staterror_Signal-Region", "Signal strength"]
    assert mock_minos.call_args[1] == {}


@mock.patch(
    "cabinetry.fit._fit_model_custom",
    return_value=fit.FitResults(
        np.asarray([1.2]), np.asarray([0.2]), ["par"], np.empty(0), 2.0
    ),
)
@mock.patch(
    "cabinetry.fit._fit_model_pyhf",
    return_value=fit.FitResults(
        np.asarray([1.1]), np.asarray([0.2]), ["par"], np.empty(0), 2.0
    ),
)
def test__fit_model(mock_pyhf, mock_custom, example_spec):
    model, data = model_utils.model_and_data(example_spec)

    # pyhf API
    fit_results = fit._fit_model(model, data)
    assert mock_pyhf.call_count == 1
    assert mock_pyhf.call_args[0][0].spec == model.spec
    assert mock_pyhf.call_args[0][1] == data
    assert mock_pyhf.call_args[1] == {
        "init_pars": None,
        "fix_pars": None,
        "minos": None,
    }
    assert np.allclose(fit_results.bestfit, [1.1])

    # pyhf API, init/fixed pars, minos
    fit_results = fit._fit_model(
        model,
        data,
        init_pars=[1.5, 2.0],
        fix_pars=[False, True],
        minos=["Signal strength"],
    )
    assert mock_pyhf.call_count == 2
    assert mock_pyhf.call_args[0][0].spec == model.spec
    assert mock_pyhf.call_args[0][1] == data
    assert mock_pyhf.call_args[1] == {
        "init_pars": [1.5, 2.0],
        "fix_pars": [False, True],
        "minos": ["Signal strength"],
    }
    assert np.allclose(fit_results.bestfit, [1.1])

    # direct iminuit
    fit_results = fit._fit_model(model, data, custom_fit=True)
    assert mock_custom.call_count == 1
    assert mock_custom.call_args[0][0].spec == model.spec
    assert mock_custom.call_args[0][1] == data
    assert mock_custom.call_args[1] == {
        "init_pars": None,
        "fix_pars": None,
        "minos": None,
    }
    assert np.allclose(fit_results.bestfit, [1.2])

    # direct iminuit, init/fixed pars, minos
    fit_results = fit._fit_model(
        model,
        data,
        init_pars=[1.5, 2.0],
        fix_pars=[False, True],
        minos=["Signal strength"],
        custom_fit=True,
    )
    assert mock_custom.call_count == 2
    assert mock_custom.call_args[0][0].spec == model.spec
    assert mock_custom.call_args[0][1] == data
    assert mock_custom.call_args[1] == {
        "init_pars": [1.5, 2.0],
        "fix_pars": [False, True],
        "minos": ["Signal strength"],
    }
    assert np.allclose(fit_results.bestfit, [1.2])


def test__run_minos(caplog):
    caplog.set_level(logging.DEBUG)

    def func_to_minimize(pars):
        # mock NLL
        return (
            np.sum(
                np.power(pars - 2 * np.ones_like(pars), 2)
                + np.power(pars - 1 * np.ones_like(pars), 4)
            )
            + pars[0]
        )

    m = iminuit.Minuit.from_array_func(
        func_to_minimize,
        [1.0, 1.0],
        name=["a", "b"],
        errordef=1,
    )
    m.migrad()
    fit._run_minos(m, ["b"], ["a", "b"])
    assert "running MINOS for b" in [rec.message for rec in caplog.records]
    assert "b = 1.5909 -0.7262 +0.4738" in [rec.message for rec in caplog.records]
    caplog.clear()

    # proper labels not known to iminuit
    m = iminuit.Minuit.from_array_func(
        func_to_minimize,
        [1.0, 1.0],
        errordef=1,
    )
    m.migrad()
    fit._run_minos(m, ["x0"], ["a", "b"])
    assert "running MINOS for a" in [rec.message for rec in caplog.records]
    assert "a = 1.3827 -0.8713 +0.5715" in [rec.message for rec in caplog.records]
    caplog.clear()

    # unknown parameter, MINOS does not run
    m = iminuit.Minuit.from_array_func(
        func_to_minimize,
        [1.0, 1.0],
        errordef=1,
    )
    m.migrad()
    fit._run_minos(m, ["x2"], ["a", "b"])
    assert [rec.message for rec in caplog.records] == [
        "parameter x2 not found in model",
        "MINOS results:",
    ]
    caplog.clear()


@mock.patch("cabinetry.model_utils.unconstrained_parameter_count", return_value=1)
def test__goodness_of_fit(mock_count, example_spec_multibin, caplog):
    caplog.set_level(logging.DEBUG)
    model, data = model_utils.model_and_data(example_spec_multibin)

    p_val = fit._goodness_of_fit(model, data, 9.964913)
    assert mock_count.call_count == 1
    assert mock_count.call_args[0][0].spec == model.spec
    assert mock_count.call_args[1] == {}
    assert "Delta NLL = 0.084185" in [rec.message for rec in caplog.records]
    assert np.allclose(p_val, 0.91926079)
    caplog.clear()


@mock.patch("cabinetry.fit._goodness_of_fit", return_value=0.1)
@mock.patch("cabinetry.fit.print_results")
@mock.patch(
    "cabinetry.fit._fit_model",
    return_value=fit.FitResults(
        np.asarray([1.0]), np.asarray([0.1]), ["par"], np.empty(0), 2.0
    ),
)
@mock.patch("cabinetry.model_utils.model_and_data", return_value=("model", "data"))
def test_fit(mock_load, mock_fit, mock_print, mock_gof, example_spec):
    # fit through pyhf.infer API
    fit_results = fit.fit(example_spec)
    assert mock_load.call_args_list == [[(example_spec,), {"asimov": False}]]
    assert mock_fit.call_args_list == [
        [("model", "data"), {"minos": None, "custom_fit": False}]
    ]
    mock_print.assert_called_once()
    assert mock_print.call_args[0][0].bestfit == [1.0]
    assert mock_print.call_args[0][0].uncertainty == [0.1]
    assert mock_print.call_args[0][0].labels == ["par"]
    assert fit_results.bestfit == [1.0]

    # Asimov fit
    fit_results = fit.fit(example_spec, asimov=True)
    assert mock_fit.call_count == 2
    assert mock_load.call_args == [(example_spec,), {"asimov": True}]
    assert fit_results.bestfit == [1.0]

    # custom fit
    fit_results = fit.fit(example_spec, custom_fit=True)
    assert mock_fit.call_count == 3
    assert mock_fit.call_args == [
        ("model", "data"),
        {"minos": None, "custom_fit": True},
    ]
    assert mock_print.call_args[0][0].bestfit == [1.0]
    assert mock_print.call_args[0][0].uncertainty == [0.1]
    assert fit_results.bestfit == [1.0]

    # parameters for MINOS
    fit_results = fit.fit(example_spec, minos=["abc"])
    assert mock_fit.call_count == 4
    assert mock_fit.call_args[1] == {"minos": ["abc"], "custom_fit": False}
    assert fit_results.bestfit == [1.0]
    fit_results = fit.fit(example_spec, minos="abc", custom_fit=True)
    assert mock_fit.call_count == 5
    assert mock_fit.call_args[1] == {"minos": ["abc"], "custom_fit": True}
    assert fit_results.bestfit == [1.0]

    # goodness-of-fit test
    fit_results_gof = fit.fit(example_spec, goodness_of_fit=True)
    assert mock_gof.call_args_list == [[("model", "data", 2.0), {}]]
    assert fit_results_gof.goodness_of_fit == 0.1


@mock.patch(
    "cabinetry.fit._fit_model",
    side_effect=[
        fit.FitResults(
            np.asarray([0.9, 1.3]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
        fit.FitResults(
            np.asarray([0.9, 0.7]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
        fit.FitResults(
            np.asarray([0.9, 1.2]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
        fit.FitResults(
            np.asarray([0.9, 0.8]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
        # for second fit with fixed parameter
        fit.FitResults(
            np.asarray([0.9, 1.2]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
        fit.FitResults(
            np.asarray([0.9, 0.8]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
    ],
)
def test_ranking(mock_fit, example_spec):
    example_spec["measurements"][0]["config"]["parameters"][0]["fixed"] = False
    bestfit = np.asarray([0.9, 1.0])
    uncertainty = np.asarray([0.02, 0.1])
    labels = ["staterror", "mu"]
    fit_results = fit.FitResults(bestfit, uncertainty, labels, np.empty(0), 0.0)
    ranking_results = fit.ranking(example_spec, fit_results)

    # correct call to fit
    expected_fix = [True, False]
    expected_inits = [[0.94956657, 1.0], [0.85043343, 1.0], [0.92, 1.0], [0.88, 1.0]]
    assert mock_fit.call_count == 4
    for i in range(4):
        assert np.allclose(
            mock_fit.call_args_list[i][1]["init_pars"], expected_inits[i]
        )
        assert np.allclose(mock_fit.call_args_list[i][1]["fix_pars"], expected_fix)
        assert mock_fit.call_args_list[i][1]["custom_fit"] is False

    # POI removed from fit results
    assert np.allclose(ranking_results.bestfit, [0.9])
    assert np.allclose(ranking_results.uncertainty, [0.02])
    assert ranking_results.labels == ["staterror"]

    # received correct mock results
    assert np.allclose(ranking_results.prefit_up, [0.3])
    assert np.allclose(ranking_results.prefit_down, [-0.3])
    assert np.allclose(ranking_results.postfit_up, [0.2])
    assert np.allclose(ranking_results.postfit_down, [-0.2])

    # fixed parameter in ranking, custom fit
    example_spec["measurements"][0]["config"]["parameters"][0]["fixed"] = True
    ranking_results = fit.ranking(example_spec, fit_results, custom_fit=True)
    # expect two calls in this ranking (and had 4 before, so 6 total): pre-fit
    # uncertainty is 0 since parameter is fixed, mock post-fit uncertainty is not 0
    assert mock_fit.call_count == 6
    assert mock_fit.call_args[1]["custom_fit"] is True
    assert np.allclose(ranking_results.prefit_up, [0.0])
    assert np.allclose(ranking_results.prefit_down, [0.0])
    assert np.allclose(ranking_results.postfit_up, [0.2])
    assert np.allclose(ranking_results.postfit_down, [-0.2])


@mock.patch(
    "cabinetry.fit._fit_model",
    side_effect=[
        fit.FitResults(
            np.asarray([0.9, 1.3]), np.asarray([0.1, 0.1]), [], np.empty(0), 8.0
        )
    ]  # nominal fit
    + [
        fit.FitResults(np.empty(0), np.empty(0), [], np.empty(0), abs(i) + 8)
        for i in np.linspace(-5, 5, 11)
    ]  # fits in scan
    + [
        fit.FitResults(
            np.asarray([0.9, 1.3]), np.asarray([0.1, 0.1]), [], np.empty(0), 2.0
        )
    ]
    * 6,  # fits for custom parameter range
)
def test_scan(mock_fit, example_spec):
    expected_scan_values = np.linspace(1.1, 1.5, 11)
    # -2 log(L) from unconstrained fit subtracted from expected NLLs
    expected_delta_nlls = np.abs(np.linspace(-5, 5, 11))

    par_name = "Signal strength"
    scan_results = fit.scan(example_spec, par_name)
    assert scan_results.name == par_name
    assert scan_results.bestfit == 1.3
    assert scan_results.uncertainty == 0.1
    assert np.allclose(scan_results.parameter_values, expected_scan_values)
    assert np.allclose(scan_results.delta_nlls, expected_delta_nlls)

    assert mock_fit.call_count == 12
    # unconstrained fit
    assert mock_fit.call_args_list[0][1] == {"custom_fit": False}
    # fits in scan
    for i, scan_val in enumerate(expected_scan_values):
        assert mock_fit.call_args_list[i + 1][1]["init_pars"] == [1.1, scan_val]
        assert mock_fit.call_args_list[i + 1][1]["fix_pars"] == [True, True]
        assert mock_fit.call_args_list[i + 1][1]["custom_fit"] is False

    # parameter range specified, custom fit
    scan_results = fit.scan(
        example_spec, par_name, par_range=(1.0, 1.5), n_steps=5, custom_fit=True
    )
    expected_custom_scan = np.linspace(1.0, 1.5, 5)
    assert np.allclose(scan_results.parameter_values, expected_custom_scan)
    assert mock_fit.call_args[1]["custom_fit"] is True

    # unknown parameter
    with pytest.raises(ValueError, match="could not find parameter abc in model"):
        fit.scan(example_spec, "abc")


def test_limit(example_spec_with_background, caplog):
    caplog.set_level(logging.DEBUG)

    # expected values for results
    observed_limit = 0.748
    expected_limit = [0.300, 0.410, 0.583, 0.833, 1.159]

    limit_results = fit.limit(example_spec_with_background)
    assert np.allclose(limit_results.observed_limit, observed_limit, rtol=1e-2)
    assert np.allclose(limit_results.expected_limit, expected_limit, rtol=1e-2)
    # compare a few CLs values
    assert np.allclose(limit_results.observed_CLs[0], 0.707447)
    assert np.allclose(
        limit_results.expected_CLs[0],
        [0.240884, 0.386364, 0.586467, 0.803260, 0.948896],
    )
    assert np.allclose(limit_results.poi_values[0], 0.152476)
    assert np.allclose(limit_results.observed_CLs[-1], 0.0)
    assert np.allclose(limit_results.expected_CLs[-1], [0.0, 0.0, 0.0, 0.0, 0.0])
    assert np.allclose(limit_results.poi_values[-1], 4.736068)
    # verify that POI values are sorted
    assert np.allclose(limit_results.poi_values, sorted(limit_results.poi_values))
    caplog.clear()

    # accesses negative POI values since upper bracket is too high
    limit_results = fit.limit(
        example_spec_with_background, bracket=[1, 5], tolerance=0.05
    )
    assert (
        "optimizer used Signal strength = -5.4721, skipping fit and setting CLs = 1"
        in [rec.message for rec in caplog.records]
    )
    assert np.allclose(limit_results.observed_limit, observed_limit, rtol=5e-2)
    assert np.allclose(limit_results.expected_limit, expected_limit, rtol=5e-2)
    caplog.clear()

    # convergence issues due to choice of bracket (needs larger bounds)
    example_spec_with_background["measurements"][0]["config"]["parameters"][0][
        "bounds"
    ] = [[0.0, 500.0]]
    fit.limit(example_spec_with_background, bracket=[10, 50])
    assert "one or more calculations did not converge, check log" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # Asimov dataset with nominal signal strength of 0
    example_spec_with_background["measurements"][0]["config"]["parameters"][0][
        "inits"
    ] = [0.0]
    limit_results = fit.limit(example_spec_with_background, asimov=True)
    assert np.allclose(limit_results.observed_limit, 0.587, rtol=1e-2)
    assert np.allclose(limit_results.expected_limit, expected_limit, rtol=2e-2)
    caplog.clear()

    # bracket with identical values
    with pytest.raises(ValueError, match="the two bracket values must not be the same"):
        fit.limit(example_spec_with_background, bracket=[3.0, 3.0])



import pathlib
from unittest import mock

import numpy as np
import pytest

from cabinetry import histo
from cabinetry import template_postprocessor


@pytest.mark.parametrize(
    "test_histo, fixed_stdev",
    [
        (
            histo.Histogram.from_arrays([1, 2, 3], [1, 2], [float("nan"), 0.2]),
            [0.0, 0.2],
        ),
        (histo.Histogram.from_arrays([1, 2, 3], [1, 2], [0.1, 0.2]), [0.1, 0.2]),
    ],
)
def test__fix_stat_unc(test_histo, fixed_stdev):
    name = "test_histo"
    template_postprocessor._fix_stat_unc(test_histo, name)
    assert np.allclose(test_histo.stdev, fixed_stdev)


def test_apply_postprocessing():
    histogram = histo.Histogram.from_arrays([1, 2, 3], [1, 1], [float("nan"), 0.2])
    name = "test_histo"
    fixed_stdev = [0.0, 0.2]
    fixed_histogram = template_postprocessor.apply_postprocessing(histogram, name)
    assert np.allclose(fixed_histogram.stdev, fixed_stdev)
    # the original histogram should be unchanged
    np.testing.assert_equal(histogram.stdev, [float("nan"), 0.2])


@mock.patch(
    "cabinetry.template_postprocessor.apply_postprocessing",
    return_value="new_histogram",
)
@mock.patch("cabinetry.histo.build_name", return_value="histo_name")
def test__get_postprocessor(mock_name, mock_apply):
    postprocessor = template_postprocessor._get_postprocessor(pathlib.Path("path"))

    region = {"Name": "region"}
    sample = {"Name": "sample"}
    systematic = {"Name": "systematic"}
    template = "Up"

    mock_original_histogram = mock.MagicMock()
    mock_new_histogram = mock.MagicMock()
    with mock.patch(
        "cabinetry.histo.Histogram.from_config", return_value=mock_original_histogram
    ) as mock_from_config:
        with mock.patch(
            "cabinetry.template_postprocessor.apply_postprocessing",
            return_value=mock_new_histogram,
        ) as mock_postprocessing:
            # execute the provided function
            postprocessor(region, sample, systematic, template)

            # check that the relevant functions were called
            assert mock_from_config.call_args_list == [
                (
                    (pathlib.Path("path"), region, sample, systematic),
                    {"modified": False, "template": template},
                )
            ]  # histogram was created

            assert mock_postprocessing.call_args_list == [
                ((mock_original_histogram, "histo_name"), {})
            ]  # postprocessing was executed

            assert mock_new_histogram.save.call_args_list == [
                ((pathlib.Path("path/histo_name_modified"),), {})
            ]  # new histogram was saved (meaning right name was obtained)


@mock.patch("cabinetry.route.apply_to_all_templates")
@mock.patch("cabinetry.template_postprocessor._get_postprocessor", return_value="func")
def test_run(mock_postprocessor, mock_apply):
    config = {"General": {"HistogramFolder": "path/"}}
    template_postprocessor.run(config)

    assert mock_postprocessor.call_args_list == [((pathlib.Path("path/"),), {})]
    assert mock_apply.call_args_list == [((config, "func"), {})]



import pathlib
from typing import List, Optional, Tuple

import awkward1 as ak
import boost_histogram as bh
import numpy as np
import uproot4 as uproot


def from_uproot(
    ntuple_paths: List[pathlib.Path],
    pos_in_file: str,
    variable: str,
    bins: np.ndarray,
    weight: Optional[str] = None,
    selection_filter: Optional[str] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Reads an ntuple with uproot, and fills a histogram with the observable.

    The paths may contain wildcards.

    Args:
        ntuple_paths (List[pathlib.Path]): list of paths to ntuples
        pos_in_file (str): name of tree within ntuple
        variable (str): variable to bin histogram in
        bins (numpy.ndarray): bin edges for histogram
        weight (Optional[str], optional): event weight to extract, defaults to None (no
            weights applied)
        selection_filter (Optional[str], optional): filter to be applied on events,
            defaults to None (no filter)

    Returns:
        Tuple[np.ndarray, np.ndarray]:
            - yield per bin
            - stat. uncertainty per bin
    """
    # concatenate the path to file and location within file with ":"
    paths_with_trees = [str(path) + ":" + pos_in_file for path in ntuple_paths]

    # determine whether the weight is a float or an expression
    # (for which a branch needs to be read)
    if weight is not None:
        try:
            float(weight)
            weight_is_expression = False
        except ValueError:
            # weight is not a float, need to evaluate the expression
            weight_is_expression = True
    else:
        # no weight specified, all weights are 1.0
        weight_is_expression = False
        weight = "1.0"

    if weight_is_expression:
        # need to read observables and weights
        array_generator = uproot.iterate(
            paths_with_trees,
            expressions=[variable, weight],
            cut=selection_filter,
        )
        obs_list = []
        weight_list = []
        for arr in array_generator:
            obs_list.append(ak.to_numpy(arr[variable]))
            weight_list.append(ak.to_numpy(arr[weight]))
        observables = np.concatenate(obs_list)
        weights = np.concatenate(weight_list)

    else:
        # only need to read the observables
        array_generator = uproot.iterate(
            paths_with_trees,
            expressions=[variable],
            cut=selection_filter,
        )
        obs_list = []
        for arr in array_generator:
            obs_list.append(ak.to_numpy(arr[variable]))
        observables = np.concatenate(obs_list)
        weights = np.ones_like(observables) * float(weight)

    yields, stdev = _bin_data(observables, weights, bins)
    return yields, stdev


def _bin_data(
    observables: np.ndarray, weights: np.ndarray, bins: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Creates a histogram from unbinned data.

    Args:
        observables (numpy.ndarray): values the histogram will be binned in
        weights (numpy.ndarray): weights to apply for each histogram entry
        bins (numpy.ndarray): bin edges for histogram

    Returns:
        Tuple[np.ndarray, np.ndarray]:
            - yield per bin
            - stat. uncertainty per bin
    """
    hist = bh.Histogram(bh.axis.Variable(bins), storage=bh.storage.Weight())
    hist.fill(observables, weight=weights)
    yields = hist.view().value
    stdev = np.sqrt(hist.view().variance)
    return yields, stdev



import logging

import numpy as np
import pytest

import cabinetry
from util import create_ntuples


@pytest.fixture
def ntuple_creator():
    return create_ntuples.run


@pytest.mark.no_cover
def test_integration(tmp_path, ntuple_creator, caplog):
    """The purpose of this integration test is to check whether the
    steps run without error and whether the fit result is as expected.
    """
    ntuple_creator(str(tmp_path))

    cabinetry_config = cabinetry.configuration.load("config_example.yml")

    # override config options to point to tmp_path
    cabinetry_config["General"]["HistogramFolder"] = str(tmp_path / "histograms")
    cabinetry_config["General"]["InputPath"] = str(tmp_path / "{SamplePaths}")

    caplog.set_level(logging.DEBUG)

    cabinetry.template_builder.create_histograms(cabinetry_config, method="uproot")
    cabinetry.template_postprocessor.run(cabinetry_config)
    workspace_path = tmp_path / "example_workspace.json"
    ws = cabinetry.workspace.build(cabinetry_config)
    cabinetry.workspace.save(ws, workspace_path)
    ws = cabinetry.workspace.load(workspace_path)
    fit_results = cabinetry.fit.fit(ws, minos="Signal_norm", goodness_of_fit=True)

    bestfit_expected = [
        1.00102289,
        0.98910429,
        1.01970061,
        0.98296235,
        1.68953317,
        -0.0880332,
        -0.32457145,
        -0.58582788,
    ]
    uncertainty_expected = [
        0.04108053,
        0.03792499,
        0.03651557,
        0.04250593,
        0.93760399,
        0.99131386,
        0.5547287,
        0.62647293,
    ]
    best_twice_nll_expected = 17.194205
    corr_mat_expected = [
        [
            1.0,
            0.06287511,
            0.01949609,
            -0.02357913,
            -0.11102759,
            0.01070953,
            0.16237686,
            0.06730786,
        ],
        [
            0.06287511,
            1.0,
            0.00909003,
            -0.00493023,
            0.14668324,
            -0.02131046,
            -0.12244755,
            -0.18872808,
        ],
        [
            0.01949609,
            0.00909003,
            1.0,
            0.06546918,
            -0.11512695,
            0.00171275,
            0.04622224,
            0.03260444,
        ],
        [
            -0.02357913,
            -0.00493023,
            0.06546918,
            1.0,
            -0.05875069,
            -0.00072121,
            -0.04292646,
            -0.02540061,
        ],
        [
            -0.11102759,
            0.14668324,
            -0.11512695,
            -0.05875069,
            1.0,
            -0.1681882,
            -0.93098096,
            -0.89116314,
        ],
        [
            0.01070953,
            -0.02131046,
            0.00171275,
            -0.00072121,
            -0.1681882,
            1.0,
            0.0532316,
            -0.13803459,
        ],
        [
            0.16237686,
            -0.12244755,
            0.04622224,
            -0.04292646,
            -0.93098096,
            0.0532316,
            1.0,
            0.94131251,
        ],
        [
            0.06730786,
            -0.18872808,
            0.03260444,
            -0.02540061,
            -0.89116314,
            -0.13803459,
            0.94131251,
            1.0,
        ],
    ]
    assert np.allclose(fit_results.bestfit, bestfit_expected)
    assert np.allclose(fit_results.uncertainty, uncertainty_expected)
    assert np.allclose(fit_results.best_twice_nll, best_twice_nll_expected)
    assert np.allclose(fit_results.corr_mat, corr_mat_expected, rtol=1e-4)
    assert np.allclose(fit_results.goodness_of_fit, 0.24679341)

    # minos result
    assert "Signal_norm                    = 1.6895 -0.9580 +0.9052" in [
        rec.message for rec in caplog.records
    ]

    # nuisance parameter ranking
    ranking_results = cabinetry.fit.ranking(ws, fit_results, custom_fit=True)
    assert np.allclose(
        ranking_results.prefit_up,
        [
            -0.10787925,
            0.14350015,
            -0.11328328,
            -0.05573699,
            -0.14964691,
            -1.67766172,
            -1.51712811,
        ],
    )
    assert np.allclose(
        ranking_results.prefit_down,
        [
            0.10910502,
            -0.15092356,
            0.1162032,
            0.05864963,
            0.17084682,
            1.41570164,
            1.15025176,
        ],
    )
    assert np.allclose(
        ranking_results.postfit_up,
        [
            -0.10503025,
            0.13865911,
            -0.10801309,
            -0.05485107,
            -0.14842401,
            -1.09991789,
            -0.89574803,
        ],
    )
    assert np.allclose(
        ranking_results.postfit_down,
        [
            0.10604367,
            -0.14654339,
            0.11081899,
            0.05779548,
            0.16930794,
            0.84262594,
            0.76490797,
        ],
    )

    # parameter scan
    scan_results = cabinetry.fit.scan(
        ws,
        "Signal_norm",
        par_range=(1.18967971, 2.18967971),
        n_steps=3,
        custom_fit=True,
    )
    # lower edge of scan is beyond normalization factor bounds specified in workspace
    assert np.allclose(
        scan_results.delta_nlls, [0.27153966, 0.0, 0.29018620], atol=5e-5
    )

    # upper limit, this calculation is slow
    limit_results = cabinetry.fit.limit(ws, bracket=[0.5, 3.5], tolerance=0.05)
    assert np.allclose(limit_results.observed_limit, 3.2295, rtol=1e-2)
    assert np.allclose(
        limit_results.expected_limit,
        [1.0464, 1.4309, 1.8968, 2.6627, 3.4603],
        rtol=1e-2,
    )



import glob
import logging
import pathlib
from typing import Any, Dict, List, Optional, Union

import numpy as np

from . import configuration
from . import fit
from . import histo
from . import model_utils
from . import tabulate
from . import template_builder


log = logging.getLogger(__name__)


def _build_figure_name(region_name: str, is_prefit: bool) -> str:
    """Constructs a file name for a figure.

    Args:
        region_name (str): name of the region shown in the figure
        is_prefit (bool): whether the figure shows the pre- or post-fit model

    Returns:
        str: name of the file the figure should be saved to
    """
    figure_name = region_name.replace(" ", "-")
    if is_prefit:
        figure_name += "_" + "prefit"
    else:
        figure_name += "_" + "postfit"
    figure_name += ".pdf"
    return figure_name


def _total_yield_uncertainty(stdev_list: List[np.ndarray]) -> np.ndarray:
    """Calculates the absolute statistical uncertainty of a stack of MC.

    Args:
        stdev_list (List[np.ndarray]): list of absolute stat. uncertainty per sample

    Returns:
        np.array: absolute stat. uncertainty of stack of samples
    """
    tot_unc = np.sqrt(np.sum(np.power(stdev_list, 2), axis=0))
    return tot_unc


def data_MC_from_histograms(
    config: Dict[str, Any],
    figure_folder: Union[str, pathlib.Path] = "figures",
    log_scale: Optional[bool] = None,
    method: str = "matplotlib",
) -> None:
    """Draws pre-fit data/MC histograms, using histograms created by cabinetry.

    The uncertainty band drawn includes only statistical uncertainties.

    Args:
        config (Dict[str, Any]): cabinetry configuration
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        log_scale (Optional[bool], optional): whether to use logarithmic vertical axis,
            defaults to None (automatically determine whether to use linear/log scale)
        method (str, optional): backend to use for plotting, defaults to "matplotlib"

    Raises:
        NotImplementedError: when trying to plot with a method that is not supported
    """
    log.info("visualizing histogram")
    histogram_folder = pathlib.Path(config["General"]["HistogramFolder"])
    for region in config["Regions"]:
        histogram_dict_list = []
        model_stdevs = []
        # loop over samples in reverse order, such that samples that appear first in the
        # config will appear at the top of the stack in the plot (they are plotted last)
        for sample in config["Samples"][::-1]:
            is_data = sample.get("Data", False)
            histogram = histo.Histogram.from_config(
                histogram_folder, region, sample, {"Name": "Nominal"}, modified=True
            )
            histogram_dict_list.append(
                {
                    "label": sample["Name"],
                    "isData": is_data,
                    "yields": histogram.yields,
                    "variable": region["Variable"],
                }
            )
            if not is_data:
                model_stdevs.append(histogram.stdev)

        figure_name = _build_figure_name(region["Name"], True)
        total_model_unc = _total_yield_uncertainty(model_stdevs)
        bin_edges = histogram.bins

        if method == "matplotlib":
            from .contrib import matplotlib_visualize

            figure_path = pathlib.Path(figure_folder) / figure_name
            matplotlib_visualize.data_MC(
                histogram_dict_list,
                total_model_unc,
                bin_edges,
                figure_path,
                log_scale=log_scale,
            )
        else:
            raise NotImplementedError(f"unknown backend: {method}")


def data_MC(
    config: Dict[str, Any],
    spec: Dict[str, Any],
    figure_folder: Union[str, pathlib.Path] = "figures",
    fit_results: Optional[fit.FitResults] = None,
    log_scale: Optional[bool] = None,
    include_table: bool = True,
    method: str = "matplotlib",
) -> None:
    """Draws pre- and post-fit data/MC histograms from a ``pyhf`` workspace.

    Args:
        config (Dict[str, Any]): cabinetry configuration
        spec (Dict[str, Any]): ``pyhf`` workspace specification
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        fit_results (Optional[fit.FitResults]): parameter configuration to use for plot,
            includes best-fit settings and uncertainties, as well as correlation matrix,
            defaults to None (then the pre-fit configuration is drawn)
        log_scale (Optional[bool], optional): whether to use logarithmic vertical axis,
            defaults to None (automatically determine whether to use linear/log scale)
        include_table (bool, optional): whether to also output a yield table, defaults
            to True
        method (str, optional): backend to use for plotting, defaults to "matplotlib"

    Raises:
        NotImplementedError: when trying to plot with a method that is not supported
    """
    model, data_combined = model_utils.model_and_data(spec, with_aux=False)

    if fit_results is not None:
        # fit results specified, draw a post-fit plot with them applied
        prefit = False
        param_values = fit_results.bestfit
        param_uncertainty = fit_results.uncertainty
        corr_mat = fit_results.corr_mat

    else:
        # no fit results specified, draw a pre-fit plot
        prefit = True
        # use pre-fit parameter values, uncertainties, and diagonal correlation matrix
        param_values = model_utils.get_asimov_parameters(model)
        param_uncertainty = model_utils.get_prefit_uncertainties(model)
        corr_mat = np.zeros(shape=(len(param_values), len(param_values)))
        np.fill_diagonal(corr_mat, 1.0)

    yields_combined = model.main_model.expected_data(
        param_values, return_by_sample=True
    )  # all channels concatenated

    # slice the yields into an array where the first index is the channel,
    # and the second index is the sample
    region_split_indices = model_utils._get_channel_boundary_indices(model)
    model_yields = np.split(yields_combined, region_split_indices, axis=1)
    data = np.split(data_combined, region_split_indices)  # data just indexed by channel

    # calculate the total standard deviation of the model prediction, index: channel
    total_stdev_model = model_utils.calculate_stdev(
        model, param_values, param_uncertainty, corr_mat
    )

    if include_table:
        # show yield table
        if prefit:
            log.info("generating pre-fit yield table")
        else:
            log.info("generating post-fit yield table")
        tabulate._yields(model, model_yields, total_stdev_model, data)

    # process channel by channel
    for i_chan, channel_name in enumerate(model.config.channels):
        histogram_dict_list = []  # one dict per region/channel

        # get the region dictionary from the config for binning / variable name
        region_dict = configuration.get_region_dict(config, channel_name)
        bin_edges = template_builder._get_binning(region_dict)
        variable = region_dict["Variable"]

        for i_sam, sample_name in enumerate(model.config.samples):
            histogram_dict_list.append(
                {
                    "label": sample_name,
                    "isData": False,
                    "yields": model_yields[i_chan][i_sam],
                    "variable": variable,
                }
            )

        # add data sample
        histogram_dict_list.append(
            {
                "label": "Data",
                "isData": True,
                "yields": data[i_chan],
                "variable": variable,
            }
        )

        if method == "matplotlib":
            from .contrib import matplotlib_visualize

            if prefit:
                figure_path = pathlib.Path(figure_folder) / _build_figure_name(
                    channel_name, True
                )
            else:
                figure_path = pathlib.Path(figure_folder) / _build_figure_name(
                    channel_name, False
                )
            matplotlib_visualize.data_MC(
                histogram_dict_list,
                np.asarray(total_stdev_model[i_chan]),
                bin_edges,
                figure_path,
                log_scale=log_scale,
            )
        else:
            raise NotImplementedError(f"unknown backend: {method}")


def correlation_matrix(
    fit_results: fit.FitResults,
    figure_folder: Union[str, pathlib.Path] = "figures",
    pruning_threshold: float = 0.0,
    method: str = "matplotlib",
) -> None:
    """Draws a correlation matrix.

    Args:
        fit_results (fit.FitResults): fit results, including correlation matrix and
            parameter labels
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        pruning_threshold (float, optional): minimum correlation for a parameter to
            have with any other parameters to not get pruned, defaults to 0.0
        method (str, optional): backend to use for plotting, defaults to "matplotlib"

    Raises:
        NotImplementedError: when trying to plot with a method that is not supported
    """
    # create a matrix that is True if a correlation is below threshold, and True on the
    # diagonal
    below_threshold = np.where(
        np.abs(fit_results.corr_mat) < pruning_threshold, True, False
    )
    np.fill_diagonal(below_threshold, True)
    # get list of booleans specifying if everything in rows/columns is below threshold
    all_below_threshold = np.all(below_threshold, axis=0)
    # get list of booleans specifying if rows/columns correspond to fixed parameter
    # (0 correlations)
    fixed_parameter = np.all(np.equal(fit_results.corr_mat, 0.0), axis=0)
    # get indices of rows/columns where everything is below threshold, or the parameter
    # is fixed
    delete_indices = np.where(np.logical_or(all_below_threshold, fixed_parameter))
    # delete rows and columns where all correlations are below threshold / parameter is
    # fixed
    corr_mat = np.delete(
        np.delete(fit_results.corr_mat, delete_indices, axis=1), delete_indices, axis=0
    )
    labels = np.delete(fit_results.labels, delete_indices)

    figure_path = pathlib.Path(figure_folder) / "correlation_matrix.pdf"
    if method == "matplotlib":
        from .contrib import matplotlib_visualize

        matplotlib_visualize.correlation_matrix(corr_mat, labels, figure_path)
    else:
        raise NotImplementedError(f"unknown backend: {method}")


def pulls(
    fit_results: fit.FitResults,
    figure_folder: Union[str, pathlib.Path] = "figures",
    exclude_list: Optional[List[str]] = None,
    method: str = "matplotlib",
) -> None:
    """Draws a pull plot of parameter results and uncertainties.

    Args:
        fit_results (fit.FitResults): fit results, including correlation matrix and
            parameter labels
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        exclude_list (Optional[List[str]], optional): list of parameters to exclude from
            plot, defaults to None (nothing excluded)
        method (str, optional): backend to use for plotting, defaults to "matplotlib"

    Raises:
        NotImplementedError: when trying to plot with a method that is not supported
    """
    figure_path = pathlib.Path(figure_folder) / "pulls.pdf"
    labels_np = np.asarray(fit_results.labels)

    if exclude_list is None:
        exclude_list = []

    # exclude fixed parameters from pull plot
    exclude_list += [
        label
        for i_np, label in enumerate(labels_np)
        if fit_results.uncertainty[i_np] == 0.0
    ]

    # exclude staterror parameters from pull plot (they are centered at 1)
    exclude_list += [label for label in labels_np if label[0:10] == "staterror_"]

    # filter out parameters
    mask = [True if label not in exclude_list else False for label in labels_np]
    bestfit = fit_results.bestfit[mask]
    uncertainty = fit_results.uncertainty[mask]
    labels_np = labels_np[mask]

    if method == "matplotlib":
        from .contrib import matplotlib_visualize

        matplotlib_visualize.pulls(bestfit, uncertainty, labels_np, figure_path)
    else:
        raise NotImplementedError(f"unknown backend: {method}")


def ranking(
    ranking_results: fit.RankingResults,
    figure_folder: Union[str, pathlib.Path] = "figures",
    max_pars: Optional[int] = None,
    method: str = "matplotlib",
) -> None:
    """Produces a ranking plot showing the impact of parameters on the POI.

    Args:
        ranking_results (fit.RankingResults): fit results, and pre- and post-fit impacts
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        max_pars (Optional[int], optional): number of parameters to include, defaults to
            None (which means all parameters are included)
        method (str, optional): backend to use for plotting, defaults to "matplotlib"

    Raises:
        NotImplementedError: when trying to plot with a method that is not supported
    """
    figure_path = pathlib.Path(figure_folder) / "ranking.pdf"

    # sort parameters by decreasing average post-fit impact
    avg_postfit_impact = (
        np.abs(ranking_results.postfit_up) + np.abs(ranking_results.postfit_down)
    ) / 2

    # get indices to sort by decreasing impact
    sorted_indices = np.argsort(avg_postfit_impact)[::-1]
    bestfit = ranking_results.bestfit[sorted_indices]
    uncertainty = ranking_results.uncertainty[sorted_indices]
    labels = np.asarray(ranking_results.labels)[sorted_indices]  # labels are list
    prefit_up = ranking_results.prefit_up[sorted_indices]
    prefit_down = ranking_results.prefit_down[sorted_indices]
    postfit_up = ranking_results.postfit_up[sorted_indices]
    postfit_down = ranking_results.postfit_down[sorted_indices]

    if max_pars is not None:
        # only keep leading parameters in ranking
        bestfit = bestfit[:max_pars]
        uncertainty = uncertainty[:max_pars]
        labels = labels[:max_pars]
        prefit_up = prefit_up[:max_pars]
        prefit_down = prefit_down[:max_pars]
        postfit_up = postfit_up[:max_pars]
        postfit_down = postfit_down[:max_pars]

    if method == "matplotlib":
        from .contrib import matplotlib_visualize

        matplotlib_visualize.ranking(
            bestfit,
            uncertainty,
            labels,
            prefit_up,
            prefit_down,
            postfit_up,
            postfit_down,
            figure_path,
        )
    else:
        raise NotImplementedError(f"unknown backend: {method}")


def templates(
    config: Dict[str, Any],
    figure_folder: Union[str, pathlib.Path] = "figures",
    method: str = "matplotlib",
) -> None:
    """Visualizes template histograms for systematic variations.

    Args:
        config (Dict[str, Any]): cabinetry configuration
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        method (str, optional): backend to use for plotting, defaults to "matplotlib"

    Raises:
        NotImplementedError: when trying to plot with a method that is not supported
    """
    log.info("visualizing systematics templates")
    histogram_folder = pathlib.Path(config["General"]["HistogramFolder"])
    figure_folder = pathlib.Path(figure_folder) / "templates"

    # could do this via the route module instead
    for region in config["Regions"]:
        for sample in config["Samples"]:
            if sample.get("Data", False):
                # skip data
                continue

            for systematic in config["Systematics"]:
                histo_name = (
                    region["Name"]
                    + "_"
                    + sample["Name"]
                    + "_"
                    + systematic["Name"]
                    + "*_modified*"
                )
                # create a list of paths to histograms matching the pattern
                variation_paths = [
                    pathlib.Path(h_name)
                    for h_name in glob.glob(str(histogram_folder / histo_name))
                ]
                # only keep up/down variations, and sort alphabetically
                # (sorting to have consistent order, and simplified debugging)
                variation_paths = sorted(
                    [v for v in variation_paths if ("Up" in v.name or "Down" in v.name)]
                )

                if len(variation_paths) == 0:
                    # no associated templates (normalization systematics)
                    continue

                # extract nominal histogram
                nominal_histo = histo.Histogram.from_config(
                    histogram_folder,
                    region,
                    sample,
                    {"Name": "Nominal"},
                )
                bins = nominal_histo.bins
                variable = region["Variable"]
                nominal = {"yields": nominal_histo.yields, "stdev": nominal_histo.stdev}

                # extract variation histograms if they exist
                up = {}
                down = {}
                for variation_path in variation_paths:
                    var_histo = histo.Histogram.from_path(variation_path)
                    var = {"yields": var_histo.yields, "stdev": var_histo.stdev}
                    if "Up" in variation_path.parts[-1]:
                        up.update(var)
                    else:
                        down.update(var)

                figure_name = (
                    region["Name"]
                    + "_"
                    + sample["Name"]
                    + "_"
                    + systematic["Name"]
                    + ".pdf"
                )
                figure_path = figure_folder / figure_name

                if method == "matplotlib":
                    from .contrib import matplotlib_visualize

                    matplotlib_visualize.templates(
                        nominal, up, down, bins, variable, figure_path
                    )

                else:
                    raise NotImplementedError(f"unknown backend: {method}")


def scan(
    scan_results: fit.ScanResults,
    figure_folder: Union[str, pathlib.Path] = "figures",
    method: str = "matplotlib",
) -> None:
    """Visualizes the results of a likelihood scan.

    Args:
        scan_results (fit.ScanResults): results of a likelihood scan
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        method (str, optional): backend to use for plotting, defaults to "matplotlib"

    Raises:
        NotImplementedError: when trying to plot with a method that is not supported
    """
    # replace [], needed for staterrors
    figure_name = (
        "scan_" + scan_results.name.replace("[", "_").replace("]", "") + ".pdf"
    )
    figure_path = pathlib.Path(figure_folder) / figure_name

    if method == "matplotlib":
        from .contrib import matplotlib_visualize

        matplotlib_visualize.scan(
            scan_results.name,
            scan_results.bestfit,
            scan_results.uncertainty,
            scan_results.parameter_values,
            scan_results.delta_nlls,
            figure_path,
        )
    else:
        raise NotImplementedError(f"unknown backend: {method}")


def limit(
    limit_results: fit.LimitResults,
    figure_folder: Union[str, pathlib.Path] = "figures",
    method: str = "matplotlib",
) -> None:
    """Visualizes observed and expected CLs values as a function of the POI.

    Args:
        limit_results (fit.LimitResults): results of upper limit determination
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        method (str, optional): backend to use for plotting, defaults to "matplotlib"

    Raises:
        NotImplementedError: when trying to plot with a method that is not supported
    """
    figure_path = pathlib.Path(figure_folder) / "limit.pdf"

    if method == "matplotlib":
        from .contrib import matplotlib_visualize

        matplotlib_visualize.limit(
            limit_results.observed_CLs,
            limit_results.expected_CLs,
            limit_results.poi_values,
            figure_path,
        )
    else:
        raise NotImplementedError(f"unknown backend: {method}")



import copy
import logging
import pathlib
from typing import Any, Dict

import numpy as np

from . import histo
from . import route


log = logging.getLogger(__name__)


def _fix_stat_unc(histogram: histo.Histogram, name: str) -> None:
    """Replaces nan stat. unc. by zero for a histogram.

    Modifies the histogram handed over in the argument.

    Args:
        histogram (cabinetry.histo.Histogram): the histogram to fix
        name (str): histogram name for logging
    """
    nan_pos = np.where(np.isnan(histogram.stdev))[0]
    if len(nan_pos) > 0:
        log.debug(f"fixing ill-defined stat. unc. for {name}")
        histogram.stdev = np.nan_to_num(histogram.stdev, nan=0.0)


def apply_postprocessing(histogram: histo.Histogram, name: str) -> histo.Histogram:
    """Returns a new histogram with post-processing applied.

    The histogram handed to the function stays unchanged. A copy of the histogram
    receives post-processing (currently only the stat. uncertainty fix) and is then
    returned.

    Args:
        histogram (cabinetry.histo.Histogram): the histogram to postprocess
        name (str): histogram name for logging

    Returns:
        cabinetry.histo.Histogram: the fixed histogram
    """
    # copy histogram to new object to leave it unchanged
    adjusted_histogram = copy.deepcopy(histogram)
    _fix_stat_unc(adjusted_histogram, name)
    return adjusted_histogram


def _get_postprocessor(histogram_folder: pathlib.Path) -> route.ProcessorFunc:
    """Returns the post-processing function to be applied to template histograms.

    Needed by ``cabinetry.route.apply_to_all_templates``. Could alternatively create a
    ``Postprocessor`` class that contains processors.

    Args:
        histogram_folder (Union[str, pathlib.Path]): folder containing histograms

    Returns:
        route.ProcessorFunc: function to apply to a template histogram
    """

    def process_template(
        region: Dict[str, Any],
        sample: Dict[str, Any],
        systematic: Dict[str, Any],
        template: str,
    ) -> None:
        """Applies post-processing to a single histogram.

        Args:
            region (Dict[str, Any]): specifying region information
            sample (Dict[str, Any]): specifying sample information
            systematic (Dict[str, Any]): specifying systematic information
            template (str): name of the template: "Nominal", "Up", "Down"
        """
        histogram = histo.Histogram.from_config(
            histogram_folder,
            region,
            sample,
            systematic,
            modified=False,
            template=template,
        )
        histogram_name = histo.build_name(region, sample, systematic, template)
        new_histogram = apply_postprocessing(histogram, histogram_name)
        histogram.validate(histogram_name)
        new_histo_path = histogram_folder / (histogram_name + "_modified")
        new_histogram.save(new_histo_path)

    return process_template


def run(config: Dict[str, Any]) -> None:
    """Applies postprocessing to all histograms.

    Args:
        config (Dict[str, Any]): cabinetry configuration
    """
    histogram_folder = pathlib.Path(config["General"]["HistogramFolder"])
    postprocessor = _get_postprocessor(histogram_folder)
    route.apply_to_all_templates(config, postprocessor)



import copy

from matplotlib.testing.compare import compare_images
import numpy as np

from cabinetry.contrib import matplotlib_visualize


def test_data_MC(tmp_path):
    fname = tmp_path / "fig.pdf"
    histo_dict_list = [
        {
            "label": "Background",
            "isData": False,
            "yields": np.asarray([12.5, 14]),
            "variable": "x",
        },
        {
            "label": "Signal",
            "isData": False,
            "yields": np.asarray([2, 5]),
            "variable": "x",
        },
        {
            "label": "Data",
            "isData": True,
            "yields": np.asarray([13, 15]),
            "variable": "x",
        },
    ]
    total_model_unc = np.sqrt([0.17, 0.29])
    bin_edges = np.asarray([1, 2, 3])
    matplotlib_visualize.data_MC(histo_dict_list, total_model_unc, bin_edges, fname)
    assert compare_images("tests/contrib/reference/data_MC.pdf", str(fname), 0) is None
    fname.unlink()  # delete figure

    histo_dict_list_log = copy.deepcopy(histo_dict_list)
    histo_dict_list_log[0]["yields"] = np.asarray([2000, 14])
    histo_dict_list_log[2]["yields"] = np.asarray([2010, 15])
    total_model_unc_log = np.asarray([50, 1.5])
    fname_log = fname.with_name(fname.stem + "_log" + fname.suffix)

    # automatic log scale
    matplotlib_visualize.data_MC(
        histo_dict_list_log, total_model_unc_log, bin_edges, fname
    )
    assert (
        compare_images("tests/contrib/reference/data_MC_log.pdf", str(fname_log), 0)
        is None
    )
    fname_log.unlink()

    # linear scale forced
    matplotlib_visualize.data_MC(
        histo_dict_list, total_model_unc, bin_edges, fname, log_scale=False
    )
    assert compare_images("tests/contrib/reference/data_MC.pdf", str(fname), 0) is None
    fname.unlink()

    # log scale forced
    matplotlib_visualize.data_MC(
        histo_dict_list_log, total_model_unc_log, bin_edges, fname, log_scale=True
    )
    assert (
        compare_images("tests/contrib/reference/data_MC_log.pdf", str(fname_log), 0)
        is None
    )
    fname_log.unlink()


def test_correlation_matrix(tmp_path):
    fname = tmp_path / "fig.pdf"
    # one parameter is below threshold so no text is shown for it on the plot
    corr_mat = np.asarray([[1.0, 0.35, 0.002], [0.35, 1.0, -0.2], [0.002, -0.2, 1.0]])
    labels = ["a", "b", "c"]
    matplotlib_visualize.correlation_matrix(corr_mat, labels, fname)
    assert (
        compare_images("tests/contrib/reference/correlation_matrix.pdf", str(fname), 0)
        is None
    )


def test_pulls(tmp_path):
    fname = tmp_path / "fig.pdf"
    bestfit = np.asarray([-0.2, 0.0, 0.1])
    uncertainty = np.asarray([0.9, 1.0, 0.7])
    labels = np.asarray(["a", "b", "c"])
    matplotlib_visualize.pulls(bestfit, uncertainty, labels, fname)
    assert compare_images("tests/contrib/reference/pulls.pdf", str(fname), 0) is None


def test_ranking(tmp_path):
    fname = tmp_path / "fig.pdf"
    bestfit = np.asarray([0.3, -0.1])
    uncertainty = np.asarray([0.8, 1.0])
    labels = np.asarray(["jet energy scale", "modeling uncertainty"])
    impact_prefit_up = np.asarray([0.5, 0.3])
    impact_prefit_down = np.asarray([-0.4, -0.3])
    impact_postfit_up = np.asarray([0.4, 0.25])
    impact_postfit_down = np.asarray([-0.3, -0.25])

    matplotlib_visualize.ranking(
        bestfit,
        uncertainty,
        labels,
        impact_prefit_up,
        impact_prefit_down,
        impact_postfit_up,
        impact_postfit_down,
        fname,
    )
    assert compare_images("tests/contrib/reference/ranking.pdf", str(fname), 0) is None


def test_templates(tmp_path):
    fname = tmp_path / "fig.pdf"
    nominal_histo = {
        "yields": np.asarray([1.0, 1.2]),
        "stdev": np.asarray([0.05, 0.06]),
    }
    up_histo = {"yields": np.asarray([1.1, 1.4]), "stdev": np.asarray([0.05, 0.07])}
    down_histo = {"yields": np.asarray([0.8, 0.9]), "stdev": np.asarray([0.06, 0.07])}
    bin_edges = np.asarray([0.0, 1.0, 2.0])
    variable = "x"

    matplotlib_visualize.templates(
        nominal_histo,
        up_histo,
        down_histo,
        bin_edges,
        variable,
        fname,
    )
    assert (
        compare_images("tests/contrib/reference/templates.pdf", str(fname), 0) is None
    )

    # only single variation specified
    matplotlib_visualize.templates(
        nominal_histo,
        up_histo,
        {},
        bin_edges,
        variable,
        fname,
    )


def test_scan(tmp_path):
    fname = tmp_path / "fig.pdf"
    par_name = "a"
    par_mle = 1.5
    par_unc = 0.2
    par_vals = np.asarray([1.1, 1.3, 1.5, 1.7, 1.9])
    par_nlls = np.asarray([4.1, 1.0, 0.0, 1.1, 3.9])

    matplotlib_visualize.scan(par_name, par_mle, par_unc, par_vals, par_nlls, fname)
    assert compare_images("tests/contrib/reference/scan.pdf", str(fname), 0) is None


def test_limit(tmp_path):
    fname = tmp_path / "fig.pdf"
    observed_CLs = np.asarray([0.31, 0.05, 0.005, 0.0001])
    expected_CLs = np.asarray(
        [
            [0.07, 0.16, 0.35, 0.62, 0.88],
            [0.0, 0.01, 0.06, 0.23, 0.56],
            [0.0, 0.0, 0.0, 0.05, 0.2],
            [0.0, 0.0, 0.0, 0.0, 0.04],
        ]
    )
    poi_values = np.asarray([0.5, 1.0, 1.5, 2.0])

    matplotlib_visualize.limit(observed_CLs, expected_CLs, poi_values, fname)
    assert compare_images("tests/contrib/reference/limit.pdf", str(fname), 0) is None



import copy
import logging

import awkward1 as ak
import numpy as np
import pyhf

from cabinetry import model_utils


def test_model_and_data(example_spec):
    model, data = model_utils.model_and_data(example_spec)
    assert model.spec["channels"] == example_spec["channels"]
    assert model.config.modifier_settings == {
        "normsys": {"interpcode": "code4"},
        "histosys": {"interpcode": "code4p"},
    }
    assert data == [475, 1.0]

    # requesting Asimov dataset
    model, data = model_utils.model_and_data(example_spec, asimov=True)
    assert data == [51.839756, 1.0]

    # without auxdata
    model, data = model_utils.model_and_data(example_spec, with_aux=False)
    assert data == [475]


def test_get_parameter_names(example_spec):
    model = pyhf.Workspace(example_spec).model()
    labels = model_utils.get_parameter_names(model)
    assert labels == ["staterror_Signal-Region", "Signal strength"]


def test_build_Asimov_data(example_spec):
    ws = pyhf.Workspace(example_spec)
    model = ws.model()
    assert model_utils.build_Asimov_data(model) == [51.839756, 1]

    # without auxdata
    assert model_utils.build_Asimov_data(model, with_aux=False) == [51.839756]

    # respect nominal settings for normfactors
    example_spec["measurements"][0]["config"]["parameters"].append(
        {"name": "Signal strength", "inits": [2.0]}
    )
    ws = pyhf.Workspace(example_spec)
    model = ws.model()
    assert model_utils.build_Asimov_data(model, with_aux=False) == [103.679512]


def test_get_asimov_parameters(example_spec, example_spec_shapefactor):
    model = pyhf.Workspace(example_spec).model()
    pars = model_utils.get_asimov_parameters(model)
    assert np.allclose(pars, [1.0, 1.0])

    model = pyhf.Workspace(example_spec_shapefactor).model()
    pars = model_utils.get_asimov_parameters(model)
    assert np.allclose(pars, [1.0, 1.0, 1.0])

    # respect nominal settings for normfactors
    example_spec["measurements"][0]["config"]["parameters"].append(
        {"name": "Signal strength", "inits": [2.0]}
    )
    model = pyhf.Workspace(example_spec).model()
    pars = model_utils.get_asimov_parameters(model)
    assert np.allclose(pars, [1.0, 2.0])


def test_get_prefit_uncertainties(
    example_spec, example_spec_multibin, example_spec_shapefactor
):
    model = pyhf.Workspace(example_spec).model()
    unc = model_utils.get_prefit_uncertainties(model)
    assert np.allclose(unc, [0.0, 0.0])  # fixed parameter and normfactor

    model = pyhf.Workspace(example_spec_multibin).model()
    unc = model_utils.get_prefit_uncertainties(model)
    assert np.allclose(unc, [0.2, 0.4, 0.0, 0.125])

    model = pyhf.Workspace(example_spec_shapefactor).model()
    unc = model_utils.get_prefit_uncertainties(model)
    assert np.allclose(unc, [0.0, 0.0, 0.0])


def test__get_channel_boundary_indices(example_spec, example_spec_multibin):
    model = pyhf.Workspace(example_spec).model()
    indices = model_utils._get_channel_boundary_indices(model)
    assert indices == []

    model = pyhf.Workspace(example_spec_multibin).model()
    indices = model_utils._get_channel_boundary_indices(model)
    assert indices == [2]

    # add extra channel to model to test three channels (two indices needed)
    three_channel_model = copy.deepcopy(example_spec_multibin)
    extra_channel = copy.deepcopy(three_channel_model["channels"][0])
    extra_channel["name"] = "region_3"
    extra_channel["samples"][0]["modifiers"][0]["name"] = "staterror_region_3"
    three_channel_model["channels"].append(extra_channel)
    three_channel_model["observations"].append({"data": [35, 8], "name": "region_3"})
    model = pyhf.Workspace(three_channel_model).model()
    indices = model_utils._get_channel_boundary_indices(model)
    assert indices == [2, 3]


def test_calculate_stdev(example_spec, example_spec_multibin):
    model = pyhf.Workspace(example_spec).model()
    parameters = np.asarray([1.05, 0.95])
    uncertainty = np.asarray([0.1, 0.1])
    corr_mat = np.asarray([[1.0, 0.2], [0.2, 1.0]])

    total_stdev = model_utils.calculate_stdev(model, parameters, uncertainty, corr_mat)
    expected_stdev = [[8.03767016]]
    assert np.allclose(ak.to_list(total_stdev), expected_stdev)

    # pre-fit
    parameters = np.asarray([1.0, 1.0])
    uncertainty = np.asarray([0.0495665682, 0.0])
    diag_corr_mat = np.diag([1.0, 1.0])
    total_stdev = model_utils.calculate_stdev(
        model, parameters, uncertainty, diag_corr_mat
    )
    expected_stdev = [[2.56951880]]  # the staterror
    assert np.allclose(ak.to_list(total_stdev), expected_stdev)

    # multiple channels, bins, staterrors
    model = pyhf.Workspace(example_spec_multibin).model()
    parameters = np.asarray([0.9, 1.05, 1.3, 0.95])
    uncertainty = np.asarray([0.1, 0.05, 0.3, 0.1])
    corr_mat = np.asarray(
        [
            [1.0, 0.1, 0.2, 0.1],
            [0.1, 1.0, 0.2, 0.3],
            [0.2, 0.2, 1.0, 0.3],
            [0.1, 0.3, 0.3, 1.0],
        ]
    )
    total_stdev = model_utils.calculate_stdev(model, parameters, uncertainty, corr_mat)
    expected_stdev = [[8.056054, 1.670629], [2.775377]]
    for i_reg in range(2):
        assert np.allclose(total_stdev[i_reg], expected_stdev[i_reg])


def test_unconstrained_parameter_count(example_spec, example_spec_shapefactor):
    model = pyhf.Workspace(example_spec).model()
    assert model_utils.unconstrained_parameter_count(model) == 1

    model = pyhf.Workspace(example_spec_shapefactor).model()
    assert model_utils.unconstrained_parameter_count(model) == 3

    # fixed parameters are skipped in counting
    example_spec_shapefactor["measurements"][0]["config"]["parameters"].append(
        {"name": "Signal strength", "fixed": True}
    )
    model = pyhf.Workspace(example_spec_shapefactor).model()
    assert model_utils.unconstrained_parameter_count(model) == 2


def test__get_parameter_index(caplog):
    caplog.set_level(logging.DEBUG)
    labels = ["a", "b", "c"]
    par_name = "b"
    assert model_utils._get_parameter_index(par_name, labels) == 1
    assert model_utils._get_parameter_index(par_name, tuple(labels)) == 1
    caplog.clear()

    assert model_utils._get_parameter_index("x", labels) == -1
    assert "parameter x not found in model" in [rec.message for rec in caplog.records]
    caplog.clear()



import logging
from typing import Any, Dict, List, Tuple, Union

import awkward1 as ak
import numpy as np
import pyhf


log = logging.getLogger(__name__)


def model_and_data(
    spec: Dict[str, Any], asimov: bool = False, with_aux: bool = True
) -> Tuple[pyhf.pdf.Model, List[float]]:
    """Returns model and data for a ``pyhf`` workspace specification.

    Args:
        spec (Dict[str, Any]): a ``pyhf`` workspace specification
        asimov (bool, optional): whether to return the Asimov dataset, defaults to False
        with_aux (bool, optional): whether to also return auxdata, defaults to True

    Returns:
        Tuple[pyhf.pdf.Model, List[float]]:
            - a HistFactory-style model in ``pyhf`` format
            - the data (plus auxdata if requested) for the model
    """
    workspace = pyhf.Workspace(spec)
    model = workspace.model(
        modifier_settings={
            "normsys": {"interpcode": "code4"},
            "histosys": {"interpcode": "code4p"},
        }
    )  # use HistFactory InterpCode=4
    if not asimov:
        data = workspace.data(model, with_aux=with_aux)
    else:
        data = build_Asimov_data(model, with_aux=with_aux)
    return model, data


def get_parameter_names(model: pyhf.pdf.Model) -> List[str]:
    """Returns the labels of all fit parameters.

    Vectors that act on one bin per vector entry (gammas) are expanded.

    Args:
        model (pyhf.pdf.Model): a HistFactory-style model in ``pyhf`` format

    Returns:
        List[str]: names of fit parameters
    """
    labels = []
    for parname in model.config.par_order:
        for i_par in range(model.config.param_set(parname).n_parameters):
            labels.append(
                f"{parname}[bin_{i_par}]"
                if model.config.param_set(parname).n_parameters > 1
                else parname
            )
    return labels


def build_Asimov_data(model: pyhf.Model, with_aux: bool = True) -> List[float]:
    """Returns the Asimov dataset (optionally with auxdata) for a model.

    Initial parameter settings for normalization factors in the workspace are treated as
    the default settings for that parameter. Fitting the Asimov dataset will recover
    these initial settings as the maximum likelihood estimate for normalization factors.
    Initial settings for other modifiers are ignored.

    Args:
        model (pyhf.Model): the model from which to construct the dataset
        with_aux (bool, optional): whether to also return auxdata, defaults to True

    Returns:
        List[float]: the Asimov dataset
    """
    asimov_data = model.expected_data(
        get_asimov_parameters(model), include_auxdata=with_aux
    ).tolist()
    return asimov_data


def get_asimov_parameters(model: pyhf.pdf.Model) -> np.ndarray:
    """Returns a list of Asimov parameter values for a model.

    For normalization factors, initial parameter settings (specified in the workspace)
    are treated as nominal settings.

    Args:
        model (pyhf.pdf.Model): model for which to extract the parameters

    Returns:
        np.ndarray: the Asimov parameters, in the same order as
        ``model.config.suggested_init()``
    """
    # create a list of parameter names, one entry per single parameter
    # (vectors like staterror expanded)
    auxdata_pars_all = []
    for parameter in model.config.auxdata_order:
        auxdata_pars_all += [parameter] * model.config.param_set(parameter).n_parameters

    # create a list of Asimov parameters (constrained parameters at the
    # best-fit value from the aux measurement, unconstrained parameters at
    # the init specified in the workspace)
    asimov_parameters = []
    for parameter in model.config.par_order:
        # indices in auxdata list that match the current parameter
        aux_indices = [i for i, par in enumerate(auxdata_pars_all) if par == parameter]
        if aux_indices:
            # pick up best-fit value from auxdata
            inits = [
                aux for i, aux in enumerate(model.config.auxdata) if i in aux_indices
            ]
        else:
            # pick up suggested inits (for normfactors)
            inits = model.config.param_set(parameter).suggested_init
        asimov_parameters += inits

    return np.asarray(asimov_parameters)


def get_prefit_uncertainties(model: pyhf.pdf.Model) -> np.ndarray:
    """Returns a list of pre-fit parameter uncertainties for a model.

    For unconstrained parameters the uncertainty is set to 0. It is also set to 0 for
    fixed parameters (similarly to how the post-fit uncertainties are defined to be 0).

    Args:
        model (pyhf.pdf.Model): model for which to extract the parameters

    Returns:
        np.ndarray: pre-fit uncertainties for the parameters, in the same order as
        ``model.config.suggested_init()``
    """
    pre_fit_unc = []  # pre-fit uncertainties for parameters
    for parameter in model.config.par_order:
        # obtain pre-fit uncertainty for constrained, non-fixed parameters
        if (
            model.config.param_set(parameter).constrained
            and not model.config.param_set(parameter).suggested_fixed
        ):
            pre_fit_unc += model.config.param_set(parameter).width()
        else:
            if model.config.param_set(parameter).n_parameters == 1:
                # unconstrained normfactor or fixed parameter, uncertainty is 0
                pre_fit_unc.append(0.0)
            else:
                # shapefactor
                pre_fit_unc += [0.0] * model.config.param_set(parameter).n_parameters
    return np.asarray(pre_fit_unc)


def _get_channel_boundary_indices(model: pyhf.pdf.Model) -> List[int]:
    """Returns indices for splitting a concatenated list of observations into channels.

    This is useful in combination with ``pyhf.pdf.Model.expected_data``, which returns
    the yields across all bins in all channels. These indices mark the positions where a
    channel begins. No index is returned for the first channel, which begins at ``[0]``.
    The returned indices can be used with ``numpy.split``.

    Args:
        model (pyhf.pdf.Model): the model that defines the channels

    Returns:
        List[int]: indices of positions where a channel begins, no index is included for
        the first bin of the first channel (which is always at ``[0]``)
    """
    # get the amount of bins per channel
    bins_per_channel = [model.config.channel_nbins[ch] for ch in model.config.channels]
    # indices of positions where a new channel starts (from the second channel onwards)
    channel_start = [sum(bins_per_channel[:i]) for i in range(1, len(bins_per_channel))]
    return channel_start


def calculate_stdev(
    model: pyhf.pdf.Model,
    parameters: np.ndarray,
    uncertainty: np.ndarray,
    corr_mat: np.ndarray,
) -> List[List[float]]:
    """Calculates the symmetrized yield standard deviation of a model.

    Args:
        model (pyhf.pdf.Model): the model for which to calculate the standard deviations
            for all bins
        parameters (np.ndarray): central values of model parameters
        uncertainty (np.ndarray): uncertainty of model parameters
        corr_mat (np.ndarray): correlation matrix

    Returns:
        List[List[float]]: list of channels, each channel is a list of standard
        deviations per bin
    """
    # indices where to split to separate all bins into regions
    region_split_indices = _get_channel_boundary_indices(model)

    # the lists up_variations and down_variations will contain the model distributions
    # with all parameters varied individually within uncertainties
    # indices: variation, channel, bin
    up_variations = []
    down_variations = []

    # calculate the model distribution for every parameter varied up and down
    # within the respective uncertainties
    for i_par in range(model.config.npars):
        # central parameter values, but one parameter varied within uncertainties
        up_pars = parameters.copy()
        up_pars[i_par] += uncertainty[i_par]
        down_pars = parameters.copy()
        down_pars[i_par] -= uncertainty[i_par]

        # total model distribution with this parameter varied up
        up_combined = model.expected_data(up_pars, include_auxdata=False)
        up_yields = np.split(up_combined, region_split_indices)
        up_variations.append(up_yields)

        # total model distribution with this parameter varied down
        down_combined = model.expected_data(down_pars, include_auxdata=False)
        down_yields = np.split(down_combined, region_split_indices)
        down_variations.append(down_yields)

    # convert to awkward arrays for further processing
    up_variations = ak.from_iter(up_variations)
    down_variations = ak.from_iter(down_variations)

    # total variance, indices are: channel, bin
    total_variance_list = [
        np.zeros(shape=(model.config.channel_nbins[ch])) for ch in model.config.channels
    ]  # list of arrays, each array has as many entries as there are bins
    total_variance = ak.from_iter(total_variance_list)

    # loop over parameters to sum up total variance
    # first do the diagonal of the correlation matrix
    for i_par in range(model.config.npars):
        symmetric_uncertainty = (up_variations[i_par] - down_variations[i_par]) / 2
        total_variance = total_variance + symmetric_uncertainty ** 2

    labels = get_parameter_names(model)
    # continue with off-diagonal contributions if there are any
    if np.count_nonzero(corr_mat - np.diag(np.ones_like(parameters))) > 0:
        # loop over pairs of parameters
        for i_par in range(model.config.npars):
            for j_par in range(model.config.npars):
                if j_par >= i_par:
                    continue  # only loop over the half the matrix due to symmetry
                corr = corr_mat[i_par, j_par]
                # an approximate calculation could be done here by requiring
                # e.g. abs(corr) > 1e-5 to continue
                if (
                    labels[i_par][0:10] == "staterror_"
                    and labels[j_par][0:10] == "staterror_"
                ):
                    continue  # two different staterrors are orthogonal, no contribution
                sym_unc_i = (up_variations[i_par] - down_variations[i_par]) / 2
                sym_unc_j = (up_variations[j_par] - down_variations[j_par]) / 2
                # factor of two below is there since loop is only over half the matrix
                total_variance = total_variance + 2 * (corr * sym_unc_i * sym_unc_j)

    # convert to standard deviation
    total_stdev = np.sqrt(total_variance)
    log.debug(f"total stdev is {total_stdev}")
    return ak.to_list(total_stdev)


def unconstrained_parameter_count(model: pyhf.pdf.Model) -> int:
    """Returns the number of unconstrained parameters in a model.

    The number is the sum of all independent parameters in a fit. A shapefactor that
    affects multiple bins enters the count once for each independent bin. Parameters
    that are set to constant are not included in the count.

    Args:
        model (pyhf.pdf.Model): model to count parameters for

    Returns:
        int: number of unconstrained parameters
    """
    n_pars = 0
    for parname in model.config.par_order:
        if (
            not model.config.param_set(parname).constrained
            and not model.config.param_set(parname).suggested_fixed
        ):
            n_pars += model.config.param_set(parname).n_parameters
    return n_pars


def _get_parameter_index(
    par_name: str, labels: Union[List[str], Tuple[str, ...]]
) -> int:
    """Returns the position of a parameter with a given name in the list of parameters.

    Useful together with ``get_parameter_names`` to find the position of a parameter
    when the name is known. If the parameter is not found, logs an error and returns a
    default value of -1.

    Args:
        par_name (str): name of parameter to find in list
        labels (Union[List[str], Tuple[str, ...]]): list or tuple with all parameter
            names in the model

    Returns:
        int: index of parameter
    """
    par_index = next((i for i, label in enumerate(labels) if label == par_name), -1)
    if par_index == -1:
        log.error(f"parameter {par_name} not found in model")
    return par_index



import logging
import os

import cabinetry


# set up log formatting and suppress verbose output from matplotlib
logging.basicConfig(
    level=logging.DEBUG, format="%(levelname)s - %(name)s - %(message)s"
)
logging.getLogger("matplotlib").setLevel(logging.WARNING)


if __name__ == "__main__":
    # check whether input data exists
    if not os.path.exists("ntuples/"):
        print("run util/create_ntuples.py to create input data")
        raise SystemExit

    # import example config file
    cabinetry_config = cabinetry.configuration.load("config_example.yml")
    cabinetry.configuration.print_overview(cabinetry_config)

    # create template histograms
    cabinetry.template_builder.create_histograms(cabinetry_config, method="uproot")

    # perform histogram post-processing
    cabinetry.template_postprocessor.run(cabinetry_config)

    # visualize systematics templates
    cabinetry.visualize.templates(cabinetry_config)

    # build a workspace and save to file
    workspace_path = "workspaces/example_workspace.json"
    ws = cabinetry.workspace.build(cabinetry_config)
    cabinetry.workspace.save(ws, workspace_path)

    # run a fit
    ws = cabinetry.workspace.load(workspace_path)
    fit_results = cabinetry.fit.fit(ws)

    # visualize pulls and correlation matrix
    cabinetry.visualize.pulls(fit_results)
    cabinetry.visualize.correlation_matrix(fit_results)

    # visualize pre- and post-fit distributions
    cabinetry.visualize.data_MC(cabinetry_config, ws)
    cabinetry.visualize.data_MC(cabinetry_config, ws, fit_results=fit_results)



import logging
from typing import Any, Dict, List

import numpy as np
import pyhf
import tabulate


log = logging.getLogger(__name__)


def _header_name(channel_name: str, i_bin: int) -> str:
    """Constructs the header name for a column in a yield table.

    Args:
        channel_name (str): name of the channel (phase space region)
        i_bin (int): index of bin in channel

    Returns:
        str: the header name to be used for the column
    """
    if i_bin == 0:
        header_name = f"{channel_name}\nbin {i_bin+1}"
    else:
        header_name = f"\nbin {i_bin+1}"
    return header_name


def _yields(
    model: pyhf.pdf.Model,
    model_yields: List[np.ndarray],
    total_stdev_model: List[List[float]],
    data: List[np.ndarray],
) -> List[Dict[str, Any]]:
    """Outputs and returns a yield table with predicted and observed yields per bin.

    Args:
        model (pyhf.pdf.Model): the model which the table corresponds to
        model_yields (List[np.ndarray]): yields per channel, sample, and bin
        total_stdev_model (List[List[float]]): total model standard deviation per
            channel and bin
        data (List[np.ndarray]): data yield per channel and bin

    Returns:
        List[Dict[str, Any]]: yield table for use with the ``tabulate`` package
    """
    table = []  # table containing all yields

    # rows for each individual sample
    for i_sam, sample_name in enumerate(model.config.samples):
        sample_dict = {"sample": sample_name}  # one dict per sample
        for i_chan, channel_name in enumerate(model.config.channels):
            for i_bin in range(model.config.channel_nbins[channel_name]):
                sample_dict.update(
                    {
                        _header_name(
                            channel_name, i_bin
                        ): f"{model_yields[i_chan][i_sam][i_bin]:.2f}"
                    }
                )
        table.append(sample_dict)

    # dicts for total model prediction and data
    total_dict = {"sample": "total"}
    data_dict = {"sample": "data"}
    for i_chan, channel_name in enumerate(model.config.channels):
        total_model = np.sum(model_yields[i_chan], axis=0)  # sum over samples
        for i_bin in range(model.config.channel_nbins[channel_name]):
            total_dict.update(
                {
                    _header_name(channel_name, i_bin): f"{total_model[i_bin]:.2f} "
                    f"\u00B1 {total_stdev_model[i_chan][i_bin]:.2f}"
                }
            )
            data_dict.update(
                {_header_name(channel_name, i_bin): f"{data[i_chan][i_bin]:.2f}"}
            )
    table += [total_dict, data_dict]

    log.info(
        "yield table:\n"
        + tabulate.tabulate(table, headers="keys", tablefmt="fancy_grid")
    )

    return table



from . import configuration  # NOQA
from . import fit  # NOQA
from . import route  # NOQA
from . import tabulate  # NOQA
from . import template_builder  # NOQA
from . import template_postprocessor  # NOQA
from . import model_utils  # NOQA
from . import visualize  # NOQA
from . import workspace  # NOQA


__version__ = "0.1.6"



from setuptools import setup

extras_require = {"contrib": ["matplotlib", "uproot3", "uproot4>=0.0.19"]}
extras_require["test"] = sorted(
    set(
        extras_require["contrib"]
        + [
            "pytest",
            "pytest-cov>=2.5.1",
            "pydocstyle",
            "check-manifest",
            "flake8",
            "flake8-bugbear",
            "flake8-import-order",
            "flake8-print",
            "mypy",
            "typeguard",
            "black==20.8b1",
        ]
    )
)
extras_require["docs"] = sorted(
    {
        "sphinx",
        "sphinx-click",
        "sphinx-copybutton",
        "sphinx-jsonschema",
        "sphinx-rtd-theme",
    }
)

extras_require["develop"] = sorted(
    set(extras_require["test"] + extras_require["docs"] + ["pre-commit", "twine"])
)
extras_require["complete"] = sorted(set(sum(extras_require.values(), [])))

setup(
    extras_require=extras_require,
)
